import math
from functools import partial

import jax
import jax.numpy as jnp
from jax.scipy.spatial.transform import Rotation

from jaxoplanet.starry import utils
from jaxoplanet.starry.core.s2fft_rotation import (
    compute_rotation_matrices as compute_rotation_matrices_s2fft,
    rotate_flms,
)
from jaxoplanet.types import Array
from jaxoplanet.utils import get_dtype_eps


def dot_rotation_matrix(ydeg, x, y, z, theta):
    """Construct a callable to apply a spherical harmonic rotation

    Args:
        ydeg (int): The order of the spherical harmonic map
        x (float): The x component of the rotation axis
        y (float): The y component of the rotation axis
        z (float): The z component of the rotation axis
        theta (float): The rotation angle in radians
    """
    try:
        ydeg = int(ydeg)
    except TypeError as e:
        raise TypeError(f"ydeg must be an integer; got {ydeg}") from e

    if theta is None:

        def do_dot(M):
            return M

        return do_dot

    if x is None and y is None:
        if z is None:
            raise ValueError("Either x, y, or z must be specified")

        return dot_rz(ydeg, theta)

    x = 0.0 if x is None else x
    y = 0.0 if y is None else y
    z = 0.0 if z is None else z

    if jnp.shape(x) != ():
        raise ValueError(f"x must be a scalar; got {jnp.shape(x)}")
    if jnp.shape(y) != ():
        raise ValueError(f"y must be a scalar; got {jnp.shape(y)}")
    if jnp.shape(z) != ():
        raise ValueError(f"z must be a scalar; got {jnp.shape(z)}")
    if jnp.shape(theta) != ():
        raise ValueError(f"theta must be a scalar; got {jnp.shape(theta)}")

    rotation_matrices = compute_rotation_matrices_s2fft(ydeg, x, y, z, theta)
    n_max = ydeg**2 + 2 * ydeg + 1

    @jax.jit
    @partial(jnp.vectorize, signature=f"({n_max})->({n_max})")
    def do_dot(M):
        """Rotate a spherical harmonic map

        Args:
            M (Array[..., n_max]): The spherical harmonic map to rotate
        """
        if M.shape[-1] != n_max:
            raise ValueError(
                f"Dimension mismatch: Input array must have shape (..., {n_max}); "
                f"got {M.shape}"
            )
        result = []
        for ell in range(ydeg + 1):
            result.append(
                M[ell * ell : ell * ell + 2 * ell + 1] @ rotation_matrices[ell]
            )
        return jnp.concatenate(result, axis=0)

    return do_dot


def full_rotation_axis_angle(
    inc: float | None, obl: float | None, theta: float | None, theta_z: float | None
):
    """Return the axis-angle representation of the full rotation of the
       spherical harmonic map

    Args:
        inc (float): map inclination
        obl (float): map obliquity
        theta (float): rotation angle about the map z-axis
        theta_z (float): rotation angle about the sky y-axis

    Returns:
        tuple: x, y, z, angle
    """
    inc = 0.0 if inc is None else inc
    obl = 0.0 if obl is None else obl
    theta = 0.0 if theta is None else theta
    theta_z = 0.0 if theta_z is None else theta_z

    f = 0.5 * math.sqrt(2)
    si = jnp.sin(inc / 2)
    ci = jnp.cos(inc / 2)

    if theta is not None and theta_z is not None:
        sp = jnp.sin(0.5 * (obl + theta + theta_z))
        cp = jnp.cos(0.5 * (obl + theta + theta_z))
        sm = jnp.sin(0.5 * (obl - theta + theta_z))
        cm = jnp.cos(0.5 * (obl - theta + theta_z))
    else:
        sp = jnp.sin(obl / 2)
        cp = jnp.cos(obl / 2)
        sm = sp
        cm = cp

    numerator1 = f * (-si * cm + ci * cp)
    numerator2 = f * (-si * sm + sp * ci)
    numerator3 = f * (si * sm + sp * ci)
    arg = si * cm + ci * cp
    denominator = jnp.sqrt(1 - 0.5 * arg**2)

    farg = f * arg
    zero_angle = jnp.allclose(farg, 1.0, atol=get_dtype_eps(farg))
    angle = jnp.where(zero_angle, 0.0, 2 * jnp.arccos(farg))

    non_zero_angle = jnp.logical_not(zero_angle)

    # this is mostly useful for the float32 case, where
    # (1 - 0.5 * arg**2) in denominator can be negative due to numerical error
    positive_arg = arg**2 < 2.0
    axis_x = jnp.where(positive_arg & non_zero_angle, numerator1 / denominator, 1.0)
    axis_y = jnp.where(positive_arg & non_zero_angle, numerator2 / denominator, 0.0)
    axis_z = jnp.where(positive_arg & non_zero_angle, numerator3 / denominator, 0.0)

    return axis_x, axis_y, axis_z, angle


def sky_projection_axis_angle(inc: float | None, obl: float | None):
    """Return the axis-angle representation of the partial rotation of the
       map due to inclination and obliquity

    Args:
        inc (float or None): map inclination
        obl (float or None): map obliquity

    Returns:
        tuple: x, y, z, angle
    """

    if obl is None and inc is None:
        return 1.0, None, None, None

    elif obl is None:
        return 1.0, None, None, inc

    elif inc is None:
        return None, None, 1.0, obl

    else:
        co = jnp.cos(obl / 2)
        so = jnp.sin(obl / 2)
        ci = jnp.cos(inc / 2)
        si = jnp.sin(inc / 2)

        denominator = jnp.sqrt(1 - ci**2 * co**2)

        # to avoid nans for the case where ci * co == 1
        denominator = jnp.where(denominator == 0.0, 1.0, denominator)

        axis_x = si * co
        axis_y = si * so
        axis_z = -so * ci

        angle = 2 * jnp.arccos(ci * co)

        arg = jnp.linalg.norm(jnp.array([axis_x, axis_y, axis_z]))
        axis_x = jnp.where(arg > 0, axis_x / denominator, 1.0)
        axis_y = jnp.where(arg > 0, axis_y / denominator, 0.0)
        axis_z = jnp.where(arg > 0, axis_z / denominator, 0.0)

        return axis_x, axis_y, axis_z, angle


def left_project(
    ydeg: int,
    inc: float | None,
    obl: float | None,
    theta: float | None,
    theta_z: float | None,
    y: Array,
):
    """R @ y

    Args:
        ydeg (int): degree of the spherical harmonic map
        inc (float or None): map inclination
        obl (float or None): map obliquity
        theta (float or None): rotation angle about the map z-axis
        theta_z (float or None): rotation angle about the sky y-axis
        x (Array): spherical harmonic map coefficients

    Returns:
        Array: rotated spherical harmonic map coefficients
    """
    m_theta = -theta if theta is not None else theta
    m_theta_z = -theta_z if theta_z is not None else theta_z

    axis_x, axis_y, axis_z, angle = sky_projection_axis_angle(inc, obl)
    y = dot_rotation_matrix(ydeg, 1.0, None, None, -0.5 * jnp.pi)(y)
    y = dot_rotation_matrix(ydeg, None, None, 1.0, m_theta)(y)
    y = dot_rotation_matrix(ydeg, axis_x, axis_y, axis_z, angle)(y)
    y = dot_rotation_matrix(ydeg, None, None, 1.0, m_theta_z)(y)
    return y


def right_project(
    ydeg: int,
    inc: float | None,
    obl: float | None,
    theta: float | None,
    theta_z: float | None,
    y: Array,
):
    """y @ R

    Args:
        ydeg (int): degree of the spherical harmonic map
        inc (float or None): map inclination
        obl (float or None): map obliquity
        theta (float or None): rotation angle about the map z-axis
        theta_z (float or None): rotation angle about the sky y-axis
        x (Array): spherical harmonic map coefficients

    Returns:
        Array: rotated spherical harmonic map coefficients
    """
    axis_x, axis_y, axis_z, angle = sky_projection_axis_angle(inc, obl)
    m_axis_x = -axis_x if axis_x is not None else axis_x
    m_axis_y = -axis_y if axis_y is not None else axis_y
    m_axis_z = -axis_z if axis_z is not None else axis_z

    y = dot_rotation_matrix(ydeg, None, None, 1.0, theta_z)(y)
    y = dot_rotation_matrix(ydeg, m_axis_x, m_axis_y, m_axis_z, angle)(y)
    y = dot_rotation_matrix(ydeg, None, None, 1.0, theta)(y)
    y = dot_rotation_matrix(ydeg, 1.0, None, None, 0.5 * jnp.pi)(y)
    return y


@partial(jax.jit, static_argnums=(0,))
def compute_rotation_matrices(ydeg, x, y, z, theta):
    # we need the axis to be a unit vector - enforce that here
    norm = jnp.sqrt(x * x + y * y + z * z)
    # handle the case where axis is (0, 0, 0)
    x = jnp.where(norm == 0.0, 0.0, x / norm)
    y = jnp.where(norm == 0.0, 0.0, y / norm)
    z = jnp.where(norm == 0.0, 0.0, z / norm)

    s = jnp.sin(theta)
    c = jnp.cos(theta)
    ra01 = x * y * (1 - c) - z * s
    ra02 = x * z * (1 - c) + y * s
    ra11 = c + y * y * (1 - c)
    ra12 = y * z * (1 - c) - x * s
    ra20 = z * x * (1 - c) - y * s
    ra21 = z * y * (1 - c) + x * s
    ra22 = c + z * z * (1 - c)

    tol = 10 * get_dtype_eps(ra22)
    cond_neg = jnp.less(jnp.abs(ra22 + 1.0), tol)
    cond_pos = jnp.less(jnp.abs(ra22 - 1.0), tol)
    cond_full = jnp.logical_or(cond_pos, cond_neg)
    sign = cond_neg.astype(int) - cond_pos.astype(int)
    norm1 = jnp.sqrt(jnp.where(cond_full, 1, ra20 * ra20 + ra21 * ra21))
    norm2 = jnp.sqrt(jnp.where(cond_full, 1, ra02 * ra02 + ra12 * ra12))
    cos_beta = ra22
    ra22_ = jnp.where(cond_full, 0.0, ra22)
    sin_beta = jnp.where(
        cond_full,
        1 + sign * ra22,
        jnp.sqrt(1 - ra22_ * ra22_),  # type: ignore
    )
    cos_gamma = jnp.where(cond_full, ra11, -ra20 / norm1)
    sin_gamma = jnp.where(cond_full, sign * ra01, ra21 / norm1)
    cos_alpha = jnp.where(cond_full, -sign * ra22, ra02 / norm2)
    sin_alpha = jnp.where(cond_full, 1 + sign * ra22, ra12 / norm2)

    return rotar(
        ydeg,
        cos_alpha,
        sin_alpha,
        cos_beta,
        sin_beta,
        cos_gamma,
        sin_gamma,
    )[1]


def rotar(ydeg, c1, s1, c2, s2, c3, s3):
    sqrt_2 = jnp.sqrt(2.0)

    D = []
    R = []

    # D[0]; R[0 ]
    D.append(jnp.ones((1, 1)))
    R.append(jnp.ones((1, 1)))
    if ydeg == 0:
        return D, R

    # D[1]
    D1_22 = 0.5 * (1 + c2)
    D1_21 = -s2 / sqrt_2
    D1_20 = 0.5 * (1 - c2)
    D1_12 = -D1_21
    D1_11 = D1_22 - D1_20
    D1_10 = D1_21
    D1_02 = D1_20
    D1_01 = D1_12
    D1_00 = D1_22
    D.append(
        jnp.array(
            [
                [D1_00, D1_01, D1_02],
                [D1_10, D1_11, D1_12],
                [D1_20, D1_21, D1_22],
            ]
        )
    )

    # R[1]
    cosag = c1 * c3 - s1 * s3
    cosamg = c1 * c3 + s1 * s3
    sinag = s1 * c3 + c1 * s3
    sinamg = s1 * c3 - c1 * s3
    R1_11 = D1_11
    R1_21 = sqrt_2 * D1_12 * c1
    R1_01 = sqrt_2 * D1_12 * s1
    R1_12 = sqrt_2 * D1_21 * c3
    R1_10 = -sqrt_2 * D1_21 * s3
    R1_22 = D1_22 * cosag - D1_20 * cosamg
    R1_20 = -D1_22 * sinag - D1_20 * sinamg
    R1_02 = D1_22 * sinag - D1_20 * sinamg
    R1_00 = D1_22 * cosag + D1_20 * cosamg
    R.append(
        jnp.array(
            [
                [R1_00, R1_01, R1_02],
                [R1_10, R1_11, R1_12],
                [R1_20, R1_21, R1_22],
            ]
        )
    )

    tol = 10 * jnp.finfo(jnp.dtype(s2)).eps
    s2_cond = jnp.less(jnp.abs(s2), tol)
    tgbet2 = jnp.where(s2_cond, s2, (1 - c2) / jnp.where(s2_cond, 1.0, s2))

    for ell in range(2, ydeg + 1):
        D_, R_ = dlmn(ell, s1, c1, c2, tgbet2, s3, c3, D)
        D.append(D_)
        R.append(R_)

    return D, R


def dlmn(ell, s1, c1, c2, tgbet2, s3, c3, D):
    iinf = 1 - ell
    isup = -iinf

    # Last row by recurrence (Eq. 19 and 20 in Alvarez Collado et al.)
    D_ = [[0 for _ in range(2 * ell + 1)] for _ in range(2 * ell + 1)]
    D_[2 * ell][2 * ell] = 0.5 * D[-1][isup + ell - 1, isup + ell - 1] * (1 + c2)
    for m in range(isup, iinf - 1, -1):
        D_[2 * ell][m + ell] = (
            -tgbet2 * jnp.sqrt((ell + m + 1) / (ell - m)) * D_[2 * ell][m + 1 + ell]
        )
    D_[2 * ell][0] = 0.5 * D[ell - 1][isup + ell - 1, -isup + ell - 1] * (1 - c2)

    # The rows of the upper quarter triangle of the D[l;m',m) matrix
    # (Eq. 21 in Alvarez Collado et al.)
    al = ell
    al1 = al - 1
    tal1 = al + al1
    ali = 1.0 / al1
    cosaux = c2 * al * al1
    for mp in range(ell - 1, -1, -1):
        amp = mp
        laux = ell + mp
        lbux = ell - mp
        aux = ali / jnp.sqrt(laux * lbux)
        cux = jnp.sqrt((laux - 1) * (lbux - 1)) * al
        for m in range(isup, iinf - 1, -1):
            am = m
            lauz = ell + m
            lbuz = ell - m
            auz = 1.0 / jnp.sqrt(lauz * lbuz)
            fact = aux * auz
            term = tal1 * (cosaux - am * amp) * D[-1][mp + ell - 1, m + ell - 1]
            if lbuz != 1 and lbux != 1:
                cuz = jnp.sqrt((lauz - 1) * (lbuz - 1))
                term = term - D[-2][mp + ell - 2, m + ell - 2] * cux * cuz
            D_[mp + ell][m + ell] = fact * term
        iinf += 1
        isup -= 1

    # The remaining elements of the D[l;m',m) matrix are calculated
    # using the corresponding symmetry relations:
    # reflection ---> ((-1)**(m-m')) D[l;m,m') = D[l;m',m), m'<=m
    # inversion ---> ((-1)**(m-m')) D[l;-m',-m) = D[l;m',m)
    sign = 1
    iinf = -ell
    isup = ell - 1
    for m in range(ell, 0, -1):
        for mp in range(iinf, isup + 1):
            D_[mp + ell][m + ell] = sign * D_[m + ell][mp + ell]
            sign *= -1
        iinf += 1
        isup -= 1

    # Inversion
    iinf = -ell
    isup = iinf
    for m in range(ell - 1, -(ell + 1), -1):
        sign = -1
        for mp in range(isup, iinf - 1, -1):
            D_[mp + ell][m + ell] = sign * D_[-mp + ell][-m + ell]
            sign *= -1
        # iinf += 1
        isup += 1

    # Compute the real rotation matrices R from the complex ones D
    R_ = [[0 for _ in range(2 * ell + 1)] for _ in range(2 * ell + 1)]
    R_[ell][ell] = D_[ell][ell]
    cosmal = c1
    sinmal = s1
    sign = -1
    root_two = jnp.sqrt(2.0)
    for mp in range(1, ell + 1):
        cosmga = c3
        sinmga = s3
        aux = root_two * D_[ell][mp + ell]
        R_[mp + ell][ell] = aux * cosmal
        R_[-mp + ell][ell] = aux * sinmal
        for m in range(1, ell + 1):
            aux = root_two * D_[m + ell][ell]
            R_[ell][m + ell] = aux * cosmga
            R_[ell][-m + ell] = -aux * sinmga
            d1 = D_[-mp + ell][-m + ell]
            d2 = sign * D_[mp + ell][-m + ell]
            cosag = cosmal * cosmga - sinmal * sinmga
            cosagm = cosmal * cosmga + sinmal * sinmga
            sinag = sinmal * cosmga + cosmal * sinmga
            sinagm = sinmal * cosmga - cosmal * sinmga
            R_[mp + ell][m + ell] = d1 * cosag + d2 * cosagm
            R_[mp + ell][-m + ell] = -d1 * sinag + d2 * sinagm
            R_[-mp + ell][m + ell] = d1 * sinag + d2 * sinagm
            R_[-mp + ell][-m + ell] = d1 * cosag - d2 * cosagm
            aux = cosmga * c3 - sinmga * s3
            sinmga = sinmga * c3 + cosmga * s3
            cosmga = aux
        sign *= -1
        aux = cosmal * c1 - sinmal * s1
        sinmal = sinmal * c1 + cosmal * s1
        cosmal = aux

    return jnp.asarray(D_), jnp.asarray(R_)


def dot_rz(deg, theta):
    """Special case for rotation only around z axis"""
    c = jnp.cos(theta)
    s = jnp.sin(theta)
    cosnt = [1.0, c]
    sinnt = [0.0, s]
    for n in range(2, deg + 1):
        cosnt.append(2.0 * cosnt[n - 1] * c - cosnt[n - 2])
        sinnt.append(2.0 * sinnt[n - 1] * c - sinnt[n - 2])

    n = 0
    cosmt = []
    sinmt = []
    for ell in range(deg + 1):
        for m in range(-ell, 0):
            cosmt.append(cosnt[-m])
            sinmt.append(-sinnt[-m])
        for m in range(ell + 1):
            cosmt.append(cosnt[m])
            sinmt.append(sinnt[m])

    n_max = deg**2 + 2 * deg + 1

    @jax.jit
    @partial(jnp.vectorize, signature=f"({n_max})->({n_max})")
    def impl(M):
        result = [0 for _ in range(n_max)]
        for ell in range(deg + 1):
            for j in range(2 * ell + 1):
                result[ell * ell + j] = (
                    M[ell * ell + j] * cosmt[ell * ell + j]
                    + M[ell * ell + 2 * ell - j] * sinmt[ell * ell + j]
                )

        return jnp.array(result, dtype=jnp.dtype(M))

    return impl


def fast_direct_left_project(ydeg, inc, obl, theta, theta_z, y):

    def euler(x, y, z, theta):
        """axis-angle to euler angles"""
        # the jnp where for theta == 0 is to avoid nans when computing grad
        axis = jnp.array([jnp.where(theta == 0.0, 1.0, x), y, z])
        _theta = jnp.where(theta == 0.0, 1.0, theta)
        axis = axis / jnp.linalg.norm(axis)
        r = Rotation.from_rotvec(axis * _theta)
        return jnp.where(theta == 0.0, jnp.array([0.0, 0.0, 0.0]), r.as_euler("zyz"))

    _x, _y, _z, angle = full_rotation_axis_angle(inc, obl, theta, theta_z)

    _axis = jnp.array([-_x, -_y, _z])
    alpha, beta, gamma = euler(*_axis, -angle)

    u = utils.C(ydeg)
    u_dag = jnp.conj(u.T)
    y_complex = jnp.array(y, dtype=jnp.complex128)
    y_d2 = utils.y1d_to_2d(ydeg, y_complex)
    y_2d_complex = (u_dag @ y_d2.T).T

    y2d_rotated_complex = rotate_flms(y_2d_complex, ydeg + 1, (alpha, beta, gamma))
    y2d_rotated_real = u @ y2d_rotated_complex.T
    y_rotated = utils.y2d_to_1d(ydeg, y2d_rotated_real.T).real
    return y_rotated



__all__ = ["ObjectStack"]

from collections.abc import Callable, Sequence
from functools import wraps
from typing import Any, Generic, TypeVar

import equinox as eqx
import jax
import jax.numpy as jnp
import jpu.numpy as jnpu
from jax.interpreters import batching
from jax.tree_util import tree_flatten

try:
    from jax.extend import linear_util as lu
except ImportError:
    from jax import linear_util as lu  # type: ignore

Obj = TypeVar("Obj")


class ObjectStack(eqx.Module, Generic[Obj]):
    """A stack of objects supporting vmapping even with different Pytree structure

    By default, functions can only be vmapped over a set of JAX objects when their Pytree
    structure matches, but this object generalizes that behavior to support a consistent
    interface that uses ``vmap`` whenever possible, falling back on a Python loop for
    variable Pytree structure.

    Args:
        objecst: A set of Pytree objects
    """

    objects: tuple[Obj, ...]
    stack: Obj | None

    def __init__(self, *objects: Obj):
        self.objects = objects

        # If all the objects have matching Pytree structure then we save a
        # stacked version that we can use for vmaps below. This allows for more
        # efficient evaluations in the case of multiple objects.
        self.stack = None
        if len(self.objects):
            spec = list(map(jax.tree_util.tree_structure, self.objects))
            if spec.count(spec[0]) == len(spec):
                self.stack = jax.tree_util.tree_map(
                    lambda *x: jnp.stack(x, axis=0), *self.objects
                )

    def __len__(self) -> int:
        return len(self.objects)

    def vmap(
        self,
        func: Callable,
        in_axes: int | None | Sequence[Any] = 0,
        out_axes: Any = 0,
    ) -> Callable:
        """Map a function over the objects in this stack

        If possible, this method will apply the appropriate ``jax.vmap`` to the input
        function, but if the Pytree structure of the objects don't match, this requires
        a loop over objects, applying the function separately to each object, and
        stacking the results.

        Args:
            func: The function to map. It's first positional argument must accept an
                object of the type ``Obj``.
            in_axes: The input axis specifications for all arguments after the first.
                The semantics should match ``jax.vmap``.
            out_axes: The output axis specifications, matching ``jax.vmap``.

        Returns:
            The vectorized version of ``func`` mapped over obejcts in this stack.
        """

        @wraps(func)
        def impl(*args):
            # First, normalize the "in_axes" argument so we always have an iterable
            if isinstance(in_axes, Sequence):
                in_axes_ = tuple(in_axes)
            else:
                in_axes_ = tuple(in_axes for _ in args)

            # If we have a "body_stack" we can just vmap and be done
            if self.stack is not None:
                return jax.vmap(func, in_axes=(0,) + in_axes_, out_axes=out_axes)(
                    self.stack, *args
                )

            # Otherwise we need to loop over the bodies and apply the function once for
            # each body

            # Here we flatten the input arguments and `in_axes` so that we don't have
            # to deal with Pytree logic for the `in_axes` ourselves below.
            args_flat, in_tree = tree_flatten(args, is_leaf=batching.is_vmappable)
            in_axes_flat = jax.api_util.flatten_axes(  # type: ignore
                "body_vmap in_axes", in_tree, in_axes_
            )

            # Then loop over the bodies and accumulate the function results
            results = []
            out_tree = None
            for n, body in enumerate(self.objects):
                f = lu.wrap_init(func)
                f, out_tree_ = flatten_func_for_object_vmap(f, in_tree, in_axes_flat, n)
                results.append(f.call_wrapped(body, *args_flat))  # type: ignore
                out_tree_ = out_tree_()  # type: ignore
                if out_tree is not None and out_tree_ != out_tree:
                    raise ValueError(
                        "Input function does not return consistent Pytree structure;\n"
                        f"expected: {out_tree}\n"
                        f"found: {out_tree_}\n"
                    )
                out_tree = out_tree_

            out_axes_flat = jax.api_util.flatten_axes(  # type: ignore
                "body_vmap out_axes", out_tree, out_axes
            )
            return out_tree.unflatten(  # type: ignore
                parts[0] if a is None else jnpu.stack(parts, axis=a)
                for a, *parts in zip(out_axes_flat, *results, strict=False)  # type: ignore
            )

        return impl


def index_helper(n, arg, axis):
    if axis is None:
        return arg
    else:
        idx = (slice(None),) * axis + (n,)
        return arg[idx]


@lu.transformation_with_aux  # type: ignore
def flatten_func_for_object_vmap(in_tree, in_axes_flat, index, body, *args_flat):
    args_indexed = (
        index_helper(index, *args)
        for args in zip(args_flat, in_axes_flat, strict=False)
    )
    ans = yield (body,) + in_tree.unflatten(args_indexed), {}
    yield tree_flatten(ans, is_leaf=batching.is_vmappable)



import equinox as eqx
import jax.numpy as jnp
import jpu.numpy as jnpu

from jaxoplanet import units
from jaxoplanet.types import Quantity
from jaxoplanet.units import unit_registry as ureg


class TransitOrbit(eqx.Module):
    """An orbit parameterized by the transit signal parameters.

    Args:
        period: Orbital periods of the planets [time unit].
        duration (Optional[Quantity]): Durations of transits [time unit].
            Either this or `speed` must be provided.
        speed (Optional[Quantity]): Speeds of the planets [length/time unit].
            Either this or `duration` must be provided.
        time_transit (Optional[Quantity]): The epochs of reference transits [time unit].
            Default is 0.
        impact_param (Optional[Quantity]): Impact parameters of the transits.
            Default is 0.
        radius_ratio (Optional[Quantity]): Ratio of planet radii to star radius.
            Default is 0.

    Raises:
        ValueError: If neither `speed` nor `duration` is provided.

    Properties:
        - shape: Returns the shape of the period, i.e. the number of planets.
        - central_radius: A quantity representing the radius of the central body.
            Its value is always 1 in this class, but required for compatibility reasons.
        - radius: An alias for radius_ratio required for compatibility reasons.
        - relative_position: The relative position of the orbiting body at a given time.
    """

    period: Quantity = units.field(units=ureg.d)
    speed: Quantity = units.field(units=1 / ureg.d)
    duration: Quantity = units.field(units=ureg.d)
    time_transit: Quantity = units.field(units=ureg.d)
    impact_param: Quantity = units.field(units=ureg.dimensionless)
    radius_ratio: Quantity = units.field(units=ureg.dimensionless)

    @units.quantity_input(
        period=ureg.d,
        duration=ureg.d,
        speed=1 / ureg.d,
        time_transit=ureg.d,
        impact_param=ureg.dimensionless,
        radius_ratio=ureg.dimensionless,
    )
    def __init__(
        self,
        *,
        period: Quantity,
        duration: Quantity | None = None,
        speed: Quantity | None = None,
        time_transit: Quantity | None = None,
        impact_param: Quantity | None = None,
        radius_ratio: Quantity | None = None,
    ):
        if duration is None:
            if speed is None:
                raise ValueError("Either 'speed' or 'duration' must be provided")
            self.period = period
            self.speed = speed
        else:
            self.period = period
            self.duration = duration

        if time_transit is None:
            self.time_transit = 0.0 * ureg.d
        else:
            self.time_transit = time_transit

        if impact_param is None:
            self.impact_param = 0.0 * ureg.dimensionless
        else:
            self.impact_param = impact_param

        if radius_ratio is None:
            self.radius_ratio = 0.0 * ureg.dimensionless
        else:
            self.radius_ratio = radius_ratio

        x2 = jnpu.square(1 + self.radius_ratio) - jnpu.square(self.impact_param)
        if duration is None:
            self.duration = 2 * jnpu.sqrt(jnpu.maximum(0, x2)) / self.speed
        else:
            self.speed = 2 * jnpu.sqrt(jnpu.maximum(0, x2)) / self.duration

    @property
    def shape(self) -> tuple[int, ...]:
        return jnp.shape(self.period)

    @property
    def radius(self) -> Quantity:
        return self.radius_ratio

    @property
    def central_radius(self) -> Quantity:
        return jnp.ones_like(self.period) * ureg.dimensionless

    @units.quantity_input(t=ureg.d, parallax=ureg.arcsec)
    def relative_position(
        self, t: Quantity, parallax: Quantity | None = None
    ) -> tuple[Quantity, Quantity, Quantity]:
        del parallax

        half_period = 0.5 * self.period
        ref_time = self.time_transit - half_period
        dt = jnpu.mod(t - ref_time, self.period) - half_period

        x = self.speed * dt
        y = jnpu.full_like(dt, self.impact_param)
        m = jnpu.fabs(dt) < 0.5 * self.duration
        z = (m * 1.0 - (~m) * 1.0) * ureg.dimensionless

        return x, y, z



__all__ = ["get_dtype_eps", "zero_safe_sqrt"]

import jax
import jax.numpy as jnp


def get_dtype_eps(x):
    return jnp.finfo(jax.dtypes.result_type(x)).eps


@jax.custom_jvp
def zero_safe_sqrt(x):
    return jnp.sqrt(x)


@zero_safe_sqrt.defjvp
def zero_safe_sqrt_jvp(primals, tangents):
    (x,) = primals
    (x_dot,) = tangents
    primal_out = jnp.sqrt(x)
    cond = jnp.less(x, 10 * get_dtype_eps(x))
    denom = jnp.where(cond, jnp.ones_like(x), x)
    tangent_out = 0.5 * x_dot * primal_out / denom
    return primal_out, tangent_out



import numpy as np
import pytest

from jaxoplanet.starry.orbit import SurfaceBody
from jaxoplanet.starry.surface import Surface
from jaxoplanet.starry.visualization import show_surface
from jaxoplanet.starry.ylm import Ylm


def test_show_map():
    _ = pytest.importorskip("matplotlib")
    y = Ylm.from_dense(np.hstack([1.0, np.random.rand(10) * 1e-1]))
    surface = Surface(y=y, inc=0.9, obl=-0.3, period=1.2, u=[0.5, 0.5])
    surface_body = SurfaceBody(period=1.0, surface=surface)

    show_surface(surface_body)
    show_surface(surface)
    show_surface(y)



__all__ = ["quantity_input"]

import inspect
from collections.abc import Callable
from functools import partial, wraps
from typing import Any

import jax
from pint import DimensionalityError

from jaxoplanet.units.registry import unit_registry


def quantity_input(
    func: Callable[..., Any] | None = None,
    *,
    _strict: bool = False,
    **kwargs: Any,
) -> Any:
    """A decorator to wrap functions that require quantities as inputs

    Please note, this is similar to the decorator of the same name from
    ``astropy.units``, but the behavior is slightly different, in ways that
    we'll try to highlight here.

    This decorator will wrap a function and check or convert the units of all
    inputs, such that the wrapped function can assume that all input units are
    correct. Note that all arguments must be specified by name (even when they)
    are positional, and variable ``*args`` and ``**kwargs`` arguments are not
    supported.

    By default, if a non-``Quantity`` is provided, it will be assumed to be in
    the correct units, and converted to a ``Quantity``. If the ``_strict`` flag
    is instead set to ``True``, inputting a non-``Quantity`` will raise a
    ``ValueError``.

    **Examples**

    The following function expects a length in meters and a time in seconds, and
    it returns a speed in meters per second:

    .. code-block:: python

        from jaxoplanet.units import unit_registry as ureg

        @units.quantity_input(a=ureg.m, b=ureg.s)
        def speed(a, b):
            return a / b

    If we call this function with a length and a time, it will work as expected:

    .. code-block:: python

        speed(1.5 * ureg.m, 0.5 * ureg.s)

    And it will also handle unit conversions:

    .. code-block:: python

        speed(1.5 * ureg.AU, 0.5 * ureg.day)  # The result will still be in m/s

    To skip validating specific inputs, you can set the unit to ``None``, or
    omit it from the decorator arguments:

    .. code-block:: python

        @units.quantity_input(x=ureg.m)  # optionally include flag=None
        def condition(x, flag):
            if flag:
                return x
            else:
                return 0.0 * x

    **JAX Pytree support**

    This decorator also supports JAX Pytrees, so you can use it to wrap
    functions with structured inputs. For example, we could rewrite the
    ``speed`` example from above as:

    .. code-block:: python

        @units.quantity_input(params={"distance": ureg.m, "time": ureg.s})
        def speed(params):
            return params["distance"] / params["time"]

    This will work with arbitrary Pytrees, as long as structure of the input
    fully matches the decorator argument. In other words, since the full Pytree
    structure must be specified, you'll need to explicitly list ``None`` for any
    Pytree nodes that you want to skip during validation:

    .. code-block:: python

        # Omitting `flag` from the decorator wouldn't work here
        @units.quantity_input(params={"x": ureg.m, "flag": None})
        def condition(params):
            if params["flag"]:
                return params["x"]
            else:
                return 0.0 * params["x"]
    """
    if func is None:
        return QuantityInput(_strict=_strict, **kwargs)
    else:
        if not callable(func):
            raise TypeError(
                "The first argument to 'quantity_input' must be a callable function, "
                "and all unit specifications must be passed as keyword arguments "
                "by name"
            )
        return QuantityInput(_strict=_strict, **kwargs)(func)


class QuantityInput:
    """This helper class implements the logic for ``quantity_input``

    Typically users should expect to interact primarily with the
    ``quantity_input`` decorator directly instead of this class, but this
    enables the use of ``quantity_input`` either as a decorator with arguments
    or as a single function call.
    """

    def __init__(
        self,
        *,
        _strict: bool = False,
        **kwargs: Any,
    ):
        self.decorator_kwargs = kwargs
        self.strict = _strict

    def __call__(self, func: Callable) -> Callable:
        signature = inspect.signature(func)
        bound_input_units = signature.bind_partial(**self.decorator_kwargs)

        input_unit_map = {}
        for param in signature.parameters.values():
            if param.kind in (
                inspect.Parameter.VAR_KEYWORD,
                inspect.Parameter.VAR_POSITIONAL,
            ):
                # We typically ignore *args and **kwargs, but if they were
                # explicitly specified, we raise an error since we don't support
                # that use case
                if param.name in bound_input_units.arguments:
                    raise TypeError(
                        "Units for general variable arguments and keyword "
                        "arguments are not supported"
                    )
            else:
                input_unit_map[param.name] = bound_input_units.arguments.get(
                    param.name, None
                )

        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            bound_args = signature.bind(*args, **kwargs)
            bound_args.apply_defaults()

            for name, value in bound_args.arguments.items():
                # If the value and the default value is None, pass through
                # ignoring units
                if value is None and signature.parameters[name].default is None:
                    continue

                unit = input_unit_map.get(name, None)
                if unit is not None:
                    bound_args.arguments[name] = jax.tree_util.tree_map(
                        partial(_apply_units, name=name, strict=self.strict),
                        value,
                        unit,
                        is_leaf=_is_quantity,
                    )

            return func(*bound_args.args, **bound_args.kwargs)

        return wrapped


def _apply_units(
    value: Any, units: Any, strict: bool = False, name: str | None = None
) -> Any:
    if units is None:
        return value
    if _is_quantity(value):
        try:
            return value.to(units)
        except DimensionalityError as e:
            raise DimensionalityError(
                e.units1,
                e.units2,
                e.dim1,
                e.dim2,
                "" if name is None else f" for input '{name}'",
            ) from None
    elif strict:
        raise ValueError("Arguments must be quantities for strict parsing")
    else:
        return unit_registry.Quantity(value, units)


def _is_quantity(x: Any) -> bool:
    return hasattr(x, "_magnitude") and hasattr(x, "_units")



import warnings

import numpy as np
import pytest

from jaxoplanet.starry.core.basis import A1, A2_inv, basis, poly_basis
from jaxoplanet.test_utils import assert_allclose

TOLERANCE = 1e-15


@pytest.mark.skip(reason="superseded by multi-precision tests")
@pytest.mark.parametrize("lmax", [10, 7, 5, 4, 3, 2, 1, 0])
def test_A1_symbolic(lmax):
    pytest.importorskip("sympy")
    expected = A1_symbolic(lmax) / (0.5 * np.sqrt(np.pi))
    calc = A1(lmax).todense()
    np.testing.assert_allclose(calc, expected, atol=5e-12)


@pytest.mark.skip(reason="superseded by multi-precision tests")
@pytest.mark.parametrize("lmax", [10, 7, 5, 4, 3, 2, 1, 0])
def test_A2_inv_symbolic(lmax):
    pytest.importorskip("sympy")
    expected = A2_inv_symbolic(lmax)
    calc = A2_inv(lmax).todense()
    np.testing.assert_allclose(calc, expected)


@pytest.mark.parametrize("lmax", [5, 4, 3, 2, 1, 0])
def test_A1(lmax):
    pytest.importorskip("mpmath")
    from jaxoplanet.starry.multiprecision import basis as mp_basis, utils

    expected = np.atleast_2d(utils.to_numpy(mp_basis.A1(lmax)))
    calc = np.atleast_2d(A1(lmax).todense())
    np.testing.assert_allclose(calc, expected)


@pytest.mark.parametrize("lmax", [15, 10, 7, 5, 4, 3, 2, 1, 0])
def test_A2_inv(lmax):
    pytest.importorskip("mpmath")
    from jaxoplanet.starry.multiprecision import basis as mp_basis, utils

    expected = np.atleast_2d(utils.to_numpy(mp_basis.A2_inv(lmax)))
    calc = np.atleast_2d(A2_inv(lmax).todense())
    np.testing.assert_allclose(calc, expected, atol=TOLERANCE)


@pytest.mark.parametrize("lmax", [15, 10, 7, 5, 4, 3, 2, 1, 0])
def test_direct_A2(lmax):
    pytest.importorskip("mpmath")
    from jaxoplanet.starry.multiprecision import basis as mp_basis, utils

    expected = np.atleast_2d(utils.to_numpy(mp_basis.A2(lmax)))
    calc = np.atleast_2d(np.linalg.inv(A2_inv(lmax).todense()))
    np.testing.assert_allclose(calc, expected, atol=TOLERANCE)


@pytest.mark.parametrize("lmax", [10, 7, 5, 4, 3, 2, 1, 0])
def test_A1_compare_starry(lmax):
    starry = pytest.importorskip("starry")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        m = starry.Map(lmax)
        expect = m.ops.A1.eval().toarray()
    calc = A1(lmax).todense()
    np.testing.assert_allclose(calc, expect, atol=5e-12)


@pytest.mark.parametrize("lmax", [10, 7, 5, 4, 3, 2, 1, 0])
def test_A2_inv_compare_starry(lmax):
    starry = pytest.importorskip("starry")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        m = starry.Map(lmax)
        A2 = m.ops.A.eval().toarray() @ m.ops.A1Inv.eval().toarray()
    inv = A2_inv(lmax)
    np.testing.assert_allclose(inv @ A2, np.eye(inv.shape[0]), atol=5e-12)


@pytest.mark.parametrize("lmax", [10, 7, 5, 4, 3, 2, 1, 0])
def test_basis_compare_starry(lmax):
    starry = pytest.importorskip("starry")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        m = starry.Map(lmax)
        expect = m.ops.A.eval().toarray()
    calc = basis(lmax).todense()
    assert_allclose(calc, expect)


def test_poly_basis():
    expect = np.array([1.0, 0.1, 0.2, 0.3, 0.01, 0.02, 0.03, 0.06, 0.09])
    calc = poly_basis(2)(0.1, 0.3, 0.2)
    assert_allclose(calc, expect)


@pytest.mark.parametrize("lmax", [10, 5, 2, 1, 0])
def test_poly_basis_compare_starry(lmax):
    starry = pytest.importorskip("starry")

    x = np.linspace(-1, 1, 100)
    y = np.linspace(-0.5, 0.5, 100)
    z = np.linspace(-0.1, 0.1, 100)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        m = starry.Map(lmax)
        expect = m.ops.pT(x, y, z)
    calc = poly_basis(lmax)(x, y, z)
    assert_allclose(calc, expect)


def A1_symbolic(lmax):
    """The sympy implementation of the A1 matrix from the starry paper"""
    import math

    import sympy as sm
    from sympy.functions.special.tensor_functions import KroneckerDelta

    x, y = sm.symbols("x y")

    def Coefficient(expression, term):
        """Return the coefficient multiplying `term` in `expression`."""
        coeff = expression.coeff(term)
        coeff = coeff.subs(sm.sqrt(1 - x**2 - y**2), 0).subs(x, 0).subs(y, 0)
        return coeff

    def ptilde(n, x, y):
        """Return the n^th term in the polynomial basis."""
        l = math.floor(math.sqrt(n))
        m = n - l * l - l
        mu = l - m
        nu = l + m
        if nu % 2 == 0:
            i = mu // 2
            j = nu // 2
            k = 0
        else:
            i = (mu - 1) // 2
            j = (nu - 1) // 2
            k = 1
        return x**i * y**j * sm.sqrt(1 - x**2 - y**2) ** k

    def A(l, m):
        """A spherical harmonic normalization constant."""
        return sm.sqrt(
            (2 - KroneckerDelta(m, 0))
            * (2 * l + 1)
            * sm.factorial(l - m)
            / (4 * sm.pi * sm.factorial(l + m))
        )

    def B(l, m, j, k):
        """Another spherical harmonic normalization constant."""
        try:
            ratio = sm.factorial((l + m + k - 1) / 2) / sm.factorial(
                (-l + m + k - 1) / 2
            )
        except ValueError:
            ratio = 0
        return (
            2**l
            * sm.factorial(m)
            / (
                sm.factorial(j)
                * sm.factorial(k)
                * sm.factorial(m - j)
                * sm.factorial(l - m - k)
            )
            * ratio
        )

    def C(p, q, k):
        """Return the binomial theorem coefficient `C`."""
        return sm.factorial(k / 2) / (
            sm.factorial(q / 2) * sm.factorial((k - p) / 2) * sm.factorial((p - q) / 2)
        )

    def Y(l, m, x, y):
        """Return the spherical harmonic of degree `l` and order `m`."""
        res = 0
        z = sm.sqrt(1 - x**2 - y**2)
        if m >= 0:
            for j in range(0, m + 1, 2):
                for k in range(0, l - m + 1, 2):
                    for p in range(0, k + 1, 2):
                        for q in range(0, p + 1, 2):
                            res += (
                                (-1) ** ((j + p) // 2)
                                * A(l, m)
                                * B(l, m, j, k)
                                * C(p, q, k)
                                * x ** (m - j + p - q)
                                * y ** (j + q)
                            )
                for k in range(1, l - m + 1, 2):
                    for p in range(0, k, 2):
                        for q in range(0, p + 1, 2):
                            res += (
                                (-1) ** ((j + p) // 2)
                                * A(l, m)
                                * B(l, m, j, k)
                                * C(p, q, k - 1)
                                * x ** (m - j + p - q)
                                * y ** (j + q)
                                * z
                            )
        else:
            for j in range(1, abs(m) + 1, 2):
                for k in range(0, l - abs(m) + 1, 2):
                    for p in range(0, k + 1, 2):
                        for q in range(0, p + 1, 2):
                            res += (
                                (-1) ** ((j + p - 1) // 2)
                                * A(l, abs(m))
                                * B(l, abs(m), j, k)
                                * C(p, q, k)
                                * x ** (abs(m) - j + p - q)
                                * y ** (j + q)
                            )
                for k in range(1, l - abs(m) + 1, 2):
                    for p in range(0, k, 2):
                        for q in range(0, p + 1, 2):
                            res += (
                                (-1) ** ((j + p - 1) // 2)
                                * A(l, abs(m))
                                * B(l, abs(m), j, k)
                                * C(p, q, k - 1)
                                * x ** (abs(m) - j + p - q)
                                * y ** (j + q)
                                * z
                            )

        return res

    def p_Y(l, m, lmax):
        ylm = Y(l, m, x, y)
        res = [ylm.subs(sm.sqrt(1 - x**2 - y**2), 0).subs(x, 0).subs(y, 0)]
        for n in range(1, (lmax + 1) ** 2):
            res.append(Coefficient(ylm, ptilde(n, x, y)))
        return res

    def A1(lmax):
        res = sm.zeros((lmax + 1) ** 2, (lmax + 1) ** 2)
        n = 0
        for l in range(lmax + 1):
            for m in range(-l, l + 1):
                res[n] = p_Y(l, m, lmax)
                n += 1
        return res

    return sm.lambdify([], A1(lmax))()


def A2_inv_symbolic(lmax):
    """The sympy implementation of the A2 matrix from the starry paper"""
    import math

    import sympy as sm

    x, y = sm.symbols("x y")

    def Coefficient(expression, term):
        """Return the coefficient multiplying `term` in `expression`."""
        coeff = expression.coeff(term)
        coeff = coeff.subs(sm.sqrt(1 - x**2 - y**2), 0).subs(x, 0).subs(y, 0)
        return coeff

    def ptilde(n, x, y):
        """Return the n^th term in the polynomial basis."""
        l = math.floor(math.sqrt(n))
        m = n - l * l - l
        mu = l - m
        nu = l + m
        if nu % 2 == 0:
            i = mu // 2
            j = nu // 2
            k = 0
        else:
            i = (mu - 1) // 2
            j = (nu - 1) // 2
            k = 1
        return x**i * y**j * sm.sqrt(1 - x**2 - y**2) ** k

    def gtilde(n, x, y):
        """Return the n^th term in the Green's basis."""
        l = math.floor(math.sqrt(n))
        m = n - l * l - l
        mu = l - m
        nu = l + m
        if nu % 2 == 0:
            I = [mu // 2]
            J = [nu // 2]
            K = [0]
            C = [(mu + 2) // 2]
        elif (l == 1) and (m == 0):
            I = [0]
            J = [0]
            K = [1]
            C = [1]
        elif (mu == 1) and (l % 2 == 0):
            I = [l - 2]
            J = [1]
            K = [1]
            C = [3]
        elif mu == 1:
            I = [l - 3, l - 1, l - 3]
            J = [0, 0, 2]
            K = [1, 1, 1]
            C = [-1, 1, 4]
        else:
            I = [(mu - 5) // 2, (mu - 5) // 2, (mu - 1) // 2]
            J = [(nu - 1) // 2, (nu + 3) // 2, (nu - 1) // 2]
            K = [1, 1, 1]
            C = [(mu - 3) // 2, -(mu - 3) // 2, -(mu + 3) // 2]

        res = 0
        for i, j, k, c in zip(I, J, K, C, strict=False):
            res += c * x**i * y**j * sm.sqrt(1 - x**2 - y**2) ** k
        return res

    def p_G(n, lmax):
        g = gtilde(n, x, y)
        res = [g.subs(sm.sqrt(1 - x**2 - y**2), 0).subs(x, 0).subs(y, 0)]
        for i in range(1, (lmax + 1) ** 2):
            res.append(Coefficient(g, ptilde(i, x, y)))
        return res

    def A2_inv(lmax):
        res = sm.zeros((lmax + 1) ** 2, (lmax + 1) ** 2)
        n = 0
        for l in range(lmax + 1):
            for _m in range(-l, l + 1):
                res[n] = p_G(n, lmax)
                n += 1
        return res

    return sm.lambdify([], A2_inv(lmax))()



from collections import defaultdict

from jaxoplanet.starry.multiprecision import mp
from jaxoplanet.starry.multiprecision.utils import fac, kron_delta

CACHED_MATRICES = defaultdict(
    lambda: {
        "R_obl": {},
        "R_inc": {},
    }
)


def get_R(name, l_max, obl=None, inc=None, cache=None):
    if cache is None:
        cache = CACHED_MATRICES

    if obl is not None and inc is not None:
        if name == "R_obl":
            if obl not in cache[l_max][name]:
                print(f"pre-computing {name}...")
                cache[l_max][name][obl] = R(l_max, (0.0, 0.0, 1.0), -obl)
            return cache[l_max][name][obl]
        elif name == "R_inc":
            if (inc, obl) not in cache[l_max][name]:
                print(f"pre-computing {name}...")
                cache[l_max][name][(inc, obl)] = R(
                    l_max, (-mp.cos(obl), -mp.sin(obl), 0.0), (0.5 * mp.pi - inc)
                )
            return cache[l_max][name][(inc, obl)]
    else:
        if name not in cache[l_max]:
            if name == "R_px":
                print(f"pre-computing {name}...")
                cache[l_max][name] = R(l_max, (1.0, 0.0, 0.0), -0.5 * mp.pi)
            elif name == "R_mx":
                print(f"pre-computing {name}...")
                cache[l_max][name] = R(l_max, (1.0, 0.0, 0.0), 0.5 * mp.pi)

        return cache[l_max][name]


def R(lmax, u, theta):

    def Dmn(l, m, n, alpha, beta, gamma):
        """Compute the (m, n) term of the Wigner D matrix."""
        sumterm = 0
        # Expression diverges when beta = 0
        if beta == 0:
            beta = 10 ** (-mp.dps)
        for k in range(l + m + 1):
            sumterm += (
                (-1) ** k
                * mp.cos(beta / 2) ** (2 * l + m - n - 2 * k)
                * mp.sin(beta / 2) ** (-m + n + 2 * k)
                / (fac(k) * fac(l + m - k) * fac(l - n - k) * fac(n - m + k))
            )

        dmn = (
            sumterm
            * mp.exp(-mp.j * (alpha * n + gamma * m))
            * (-1) ** (n + m)
            * mp.sqrt(fac(l - m) * fac(l + m) * fac(l - n) * fac(l + n))
        )

        return dmn

    def D(l, alpha, beta, gamma):
        res = mp.zeros(2 * l + 1, 2 * l + 1)
        for m in range(-l, l + 1):
            for n in range(-l, l + 1):
                res[m + l, n + l] = Dmn(l, n, m, alpha, beta, gamma)
        return res

    def Umn(l, m, n):
        """Compute the (m, n) term of the transformation
        matrix from complex to real Ylms."""
        if n < 0:
            term1 = mp.j
        elif n == 0:
            term1 = mp.sqrt(2) / 2
        else:
            term1 = 1
        if (m > 0) and (n < 0) and (n % 2 == 0):
            term2 = -1
        elif (m > 0) and (n > 0) and (n % 2 != 0):
            term2 = -1
        else:
            term2 = 1
        return term1 * term2 * 1 / mp.sqrt(2) * (kron_delta(m, n) + kron_delta(m, -n))

    def U(l):
        """Compute the U transformation matrix."""
        res = mp.zeros(2 * l + 1, 2 * l + 1)
        for m in range(-l, l + 1):
            for n in range(-l, l + 1):
                res[m + l, n + l] = Umn(l, m, n)
        return res

    def rot_matrix(l, alpha, beta, gamma):
        """Return the rotation matrix for a single degree `l`."""
        res = mp.zeros(2 * l + 1, 2 * l + 1)
        if l == 0:
            res[0, 0] = 1
            return res
        foo = ((U(l) ** -1) * D(l, alpha, beta, gamma) * U(l)).apply(mp.re)
        for m in range(2 * l + 1):
            for n in range(2 * l + 1):
                if abs(foo[m, n]) < 10 ** (-mp.dps):
                    res[m, n] = 0
                else:
                    res[m, n] = foo[m, n]
        return res

    def axis_angle_to_euler(u1, u2, u3, theta):
        """Axis-angle rotation matrix."""

        tol = 1e-20
        if theta == 0:
            theta = tol
        if u1 == 0 and u2 == 0:
            u1 = tol
            u2 = tol

        # Elements of the transformation matrix
        costheta = mp.cos(theta)
        sintheta = mp.sin(theta)
        RA01 = u1 * u2 * (1 - costheta) - u3 * sintheta
        RA02 = u1 * u3 * (1 - costheta) + u2 * sintheta
        RA11 = costheta + u2 * u2 * (1 - costheta)
        RA12 = u2 * u3 * (1 - costheta) - u1 * sintheta
        RA20 = u3 * u1 * (1 - costheta) - u2 * sintheta
        RA21 = u3 * u2 * (1 - costheta) + u1 * sintheta
        RA22 = costheta + u3 * u3 * (1 - costheta)

        # Determine the Euler angles
        if (RA22 < -1) and (RA22 > -1):
            cosbeta = -1
            sinbeta = 0
            cosgamma = RA11
            singamma = RA01
            cosalpha = 1
            sinalpha = 0
        elif (RA22 < 1) and (RA22 > 1):
            cosbeta = 1
            sinbeta = 0
            cosgamma = RA11
            singamma = -RA01
            cosalpha = 1
            sinalpha = 0
        else:
            cosbeta = RA22
            sinbeta = mp.sqrt(1 - cosbeta**2)
            norm1 = mp.sqrt(RA20 * RA20 + RA21 * RA21)
            norm2 = mp.sqrt(RA02 * RA02 + RA12 * RA12)
            cosgamma = -RA20 / norm1
            singamma = RA21 / norm1
            cosalpha = RA02 / norm2
            sinalpha = RA12 / norm2
        alpha = mp.atan2(sinalpha, cosalpha)
        beta = mp.atan2(sinbeta, cosbeta)
        gamma = mp.atan2(singamma, cosgamma)

        return alpha, beta, gamma

    u = mp.matrix(u)
    u = u / mp.norm(u, p=2)
    alpha, beta, gamma = axis_angle_to_euler(*u, theta)
    blocks = [rot_matrix(l, alpha, beta, gamma) for l in range(lmax + 1)]

    return blocks


def dot_rotation_matrix(ydeg, u, theta, rotation_matrices=None):
    if rotation_matrices is None:
        rotation_matrices = R(ydeg, u, theta)

    def do_dot(M):
        if theta is not None:
            if mp.absmax(theta) == 0:
                return M
        result = []
        for l in range(ydeg + 1):
            if l == 0:
                result.append(mp.matrix([1]))
            else:
                result.append(
                    mp.matrix(M[l * l : l * l + 2 * l + 1]).T @ rotation_matrices[l]
                )
        S = []
        for _s in result:
            for __s in _s:
                S.append(__s)

        return mp.matrix(S)

    return do_dot


def left_project(deg, inc, obl, theta, theta_z, x):
    if theta != 0:
        x = dot_rotation_matrix(deg, (1.0, 0.0, 0.0), -0.5 * mp.pi)(x)
        x = dot_rotation_matrix(deg, (0.0, 0.0, 1.0), -theta)(x)
        x = dot_rotation_matrix(deg, (1.0, 0.0, 0.0), 0.5 * mp.pi)(x)

    if obl != 0:
        x = dot_rotation_matrix(deg, (0.0, 0.0, 1.0), -obl)(x)

    x = dot_rotation_matrix(
        deg, (-mp.cos(obl), -mp.sin(obl), 0.0), (0.5 * mp.pi - inc)
    )(x)

    if theta_z != 0:
        x = dot_rotation_matrix(deg, 0, 0, 1.0, -theta_z)(x)
    return x.T


def dot_rz(deg, theta):
    """Special case for rotation only around z axis"""
    c = mp.cos(-theta)
    s = mp.sin(-theta)
    cosnt = [1.0, c]
    sinnt = [0.0, s]
    for n in range(2, deg + 1):
        cosnt.append(2.0 * cosnt[n - 1] * c - cosnt[n - 2])
        sinnt.append(2.0 * sinnt[n - 1] * c - sinnt[n - 2])

    n = 0
    cosmt = []
    sinmt = []
    for ell in range(deg + 1):
        for m in range(-ell, 0):
            cosmt.append(cosnt[-m])
            sinmt.append(-sinnt[-m])
        for m in range(ell + 1):
            cosmt.append(cosnt[m])
            sinmt.append(sinnt[m])

    n_max = deg**2 + 2 * deg + 1

    def impl(M):
        result = [0 for _ in range(n_max)]
        for ell in range(deg + 1):
            for j in range(2 * ell + 1):
                result[ell * ell + j] = (
                    M[ell * ell + j] * cosmt[ell * ell + j]
                    + M[ell * ell + 2 * ell - j] * sinmt[ell * ell + j]
                )

        return mp.matrix(result)

    return impl



from collections.abc import Iterable

import equinox as eqx
import jax
import jax.numpy as jnp
from jax.scipy.spatial.transform import Rotation

from jaxoplanet import starry
from jaxoplanet.starry.core.basis import A1, U, poly_basis
from jaxoplanet.starry.core.polynomials import Pijk
from jaxoplanet.starry.core.rotation import full_rotation_axis_angle, left_project
from jaxoplanet.starry.utils import ortho_grid
from jaxoplanet.starry.ylm import Ylm
from jaxoplanet.types import Array, Quantity


class Surface(eqx.Module):
    """Surface map object.

    Args:
        y (Optional(:py:class:`~jaxoplanet.starry.ylm.Ylm`)) Ylm object containing the
            spherical harmonic expansion of the map. Defaults to a uniform map with
            amplitude 1.0.
        inc (Optional[Quantity]): inclination of the map relative to line of sight.
            Defaults to pi/2 [angular unit].
        obl (Optional[Quantity]): obliquity of the map [angular unit]. Defaults to None.
        u (Optional[Array]): polynomial limb-darkening coefficients of the map.
        period (Optional[Quantity]): rotation period of the map [time unit]. Defaults to
            None.
        amplitude (Optional[float]): amplitude of the map; this quantity is proportional
            to the luminosity of the map and multiplies all flux-related observables.
            Defaults to 1.0.
        normalize (Optional(bool)): whether to normalize the coefficients of the
            spherical harmonics. If True, Ylm is normalized and the amplitude of the map
            is set to y[(0, 0)]. Defaults to True.
        phase (Optional[float]): initial phase of the map rotation around the polar
            axis. Defaults to 0.0.

    Example:

        .. code-block:: python

            import numpy as np
            import jax
            from jaxoplanet.starry.visualization import show_surface
            from jaxoplanet.starry.surface import Surface
            from jaxoplanet.starry.ylm import Ylm

            jax.config.update("jax_enable_x64", True)

            np.random.seed(30)
            y = Ylm.from_dense(np.random.rand(20))
            m = Surface(y=y, u=[0.5, 0.1], inc=0.9, obl=-0.3)
            show_surface(m)
    """

    y: Ylm
    """:py:class:`~starry.ylm.Ylm` object representing the spherical harmonic expansion
        of the map"""

    _inc: Array | None
    """Inclination of the map in radians. None if seen from the pole."""

    _obl: Array | None
    """Obliquity of the map in radians. None if no obliquity."""

    u: tuple[Array, ...]
    """Tuple of limb darkening coefficients."""

    period: Array | None
    """Rotation period of the map in days (attribute subject to change). None if not
    rotating."""

    amplitude: Array
    """Amplitude of the map, a quantity proportional to map luminosity."""

    normalize: bool
    """Boolean to specify whether the Ylm coefficients should be normalized"""

    phase: Array
    """Initial phase of the map rotation around polar axis"""

    radius: Array
    """Radius of the map in solar radii"""

    shear: Array
    """Differential rotation shear of the map"""

    def __init__(
        self,
        *,
        y: Ylm | None = None,
        inc: Quantity | None = 0.5 * jnp.pi,
        obl: Quantity | None = None,
        u: Iterable[Array] = (),
        period: Quantity | None = None,
        amplitude: Array = 1.0,
        normalize: bool = True,
        phase: Array = 0.0,
        radius: Array = 1.0,
        shear: Array = None,
    ):
        if y is None:
            y = Ylm()

        if normalize:
            amplitude = jnp.array(y[(0, 0)], float)
            y = Ylm(data=y.data).normalize()

        self.y = y
        self._inc = inc
        self._obl = obl
        self.u = tuple(u)
        self.period = period
        self.amplitude = amplitude
        self.normalize = normalize
        self.phase = phase
        self.radius = radius
        self.shear = shear

    @property
    def inc(self):
        return self._inc if self._inc is not None else 0.0

    @inc.setter
    def inc(self, value):
        self._inc = value

    @property
    def obl(self):
        return self._obl if self._obl is not None else 0.0

    @obl.setter
    def obl(self, value):
        self._obl = value

    @property
    def veq(self):
        """Equatorial velocity of the map in Rsun/day."""
        return 2 * jnp.pi * self.radius / self.period

    def _poly_basis(self, rv=False):
        if rv:
            return jax.jit(poly_basis(self.deg + self.vdeg))
        else:
            return jax.jit(poly_basis(self.deg))

    @property
    def udeg(self) -> int:
        """Order of the polynomial limb darkening."""
        return len(self.u)

    @property
    def ydeg(self) -> int:
        return self.y.deg

    @property
    def vdeg(self) -> int:
        if self.shear is not None:
            return
        else:
            return 1

    @property
    def deg(self):
        """Total degree of the spherical harmonic expansion (``udeg + ydeg``)."""
        return self.ydeg + self.udeg

    def _intensity(self, x, y, z, theta=None, rv=False):
        pT = self._poly_basis(rv)(x, y, z)
        Ry = left_project(self.ydeg, self.inc, self.obl, theta, 0.0, self.y.todense())
        A1Ry = A1(self.ydeg).todense() @ Ry
        p_y = Pijk.from_dense(A1Ry, degree=self.ydeg)
        u = jnp.array([1, *self.u])
        p_u = Pijk.from_dense(u @ U(self.udeg), degree=self.udeg)
        p = p_y * p_u

        if rv:
            y_rv = starry.doppler.rv_map_expansion(
                inc=self._inc, obl=self._obl, veq=self.veq, alpha=None
            )
            p_rv = Pijk.from_dense(
                jax.experimental.sparse.BCOO.from_scipy_sparse(A1(self.vdeg)).todense()
                @ y_rv
            )
            p = p_y * p_u * p_rv

        return pT @ p.todense() * self.amplitude

    # @partial(jax.jit, static_argnames=("res",))
    def render(self, theta: float | None = None, res: int = 400, rv: bool = False):
        """Returns the intensity map projected onto the x-y plane (sky).

        Args:
            theta (float, optional): rotation angle of the map. Defaults to 0.0.
            res (int, optional): resolution of the render. Defaults to 400.

        Returns:
            ArrayLike: square 2D array representing the intensity map
            (with nans outside the map disk).
        """
        _, xyz = ortho_grid(res)
        intensity = self._intensity(*xyz, theta=theta, rv=rv)
        return jnp.reshape(intensity, (res, res))

    def intensity(self, lat: float, lon: float):
        """Returns the intensity of the map at a given latitude and longitude.

        Args:
            lat (float): latitude in the rest frame of the map
            lon (float): longitude in the rest frame of the map

        Returns:
            float: intensity of the map at the given latitude and longitude
        """
        lon = lon + jnp.pi / 2  # convention, 0 lon faces the observer
        lat = jnp.pi / 2 - lat  # convention, latitude 0 is equator
        x = jnp.sin(lat) * jnp.cos(lon)
        y = jnp.sin(lat) * jnp.sin(lon)
        z = jnp.cos(lat) * jnp.ones_like(x)

        axis = full_rotation_axis_angle(self.inc - jnp.pi / 2, self.obl, 0.0, 0.0)
        axis = jnp.array(axis[0:3]) * axis[-1]
        rotation = Rotation.from_rotvec(axis)
        x, y, z = rotation.apply(jnp.array([x, y, z]).T).T

        return self._intensity(x, y, z)

    def rotational_phase(self, time: Array) -> Array | None:
        """Returns the rotational phase of the map at a given time.

        Args:
            time (ArrayLike): time in same units as the period

        Returns:
            ArrayLike: rotational phase of the map at the given time
        """
        if self.period is None:
            return None
        else:
            return 2 * jnp.pi * time / self.period + self.phase



__all__ = ["LightCurveOrbit"]

from typing import Protocol

from jaxoplanet.types import Quantity


class LightCurveOrbit(Protocol):
    """An interface for orbits that can be used to compute light curves"""

    @property
    def shape(self) -> tuple[int, ...]: ...

    @property
    def radius(self) -> Quantity: ...

    def relative_position(self, t: Quantity) -> tuple[Quantity, Quantity, Quantity]: ...

    @property
    def central_radius(self) -> Quantity: ...



import jax
import numpy as np
import pytest

from jaxoplanet.light_curves import limb_dark_light_curve
from jaxoplanet.light_curves.emission import light_curve as emission_light_curve
from jaxoplanet.orbits import keplerian
from jaxoplanet.starry import Surface, Ylm
from jaxoplanet.starry.light_curves import light_curve, surface_light_curve
from jaxoplanet.starry.orbit import SurfaceSystem
from jaxoplanet.test_utils import assert_allclose
from jaxoplanet.units import unit_registry as ureg


@pytest.mark.parametrize("deg", [2, 5, 10])
@pytest.mark.parametrize("u", [[], [0.1], [0.2, 0.1]])
def test_compare_starry_limb_dark(deg, u):
    starry = pytest.importorskip("starry")
    starry.config.lazy = False

    # map
    inc = np.pi / 2
    np.random.seed(deg)
    y = Ylm.from_dense(np.random.randn((deg + 1) ** 2), normalize=True)
    map = Surface(y=y, u=u, inc=inc)

    # occultor
    yo = np.linspace(-3, 3, 1000)
    ro = 0.01
    zo = 2.0
    xo = 0.0

    # starry
    ms = starry.Map(ydeg=deg, udeg=len(u), inc=np.rad2deg(inc))
    ms[:, :] = y.todense()
    if len(u) > 0:
        ms[1:] = u
    expected = ms.flux(xo=xo, yo=yo, ro=ro, zo=zo)

    # jaxoplanet
    calc = jax.vmap(surface_light_curve, in_axes=(None, None, None, 0, None, None))(
        map, ro, xo, yo, zo, 0.0
    )

    assert_allclose(calc, expected)


@pytest.mark.parametrize("r", [0.01, 0.1, 1.0, 10.0, 100.0])
def test_flux(r, l_max=5, order=500):
    pytest.importorskip("mpmath")
    from jaxoplanet.starry.multiprecision import flux as mp_flux

    # We know that these are were the errors are the highest
    b = 1 - r if r < 1 else r

    n = (l_max + 1) ** 2
    expect = np.zeros(n)
    calc = np.zeros(n)
    ys = np.eye(n, dtype=np.float64)
    ys[:, 0] = 1.0

    @jax.jit
    def light_curve(y):
        surface = Surface(y=Ylm.from_dense(y, normalize=False), normalize=False)
        return surface_light_curve(surface, y=b, z=10.0, r=r, order=order)

    for i, y in enumerate(ys):
        expect[i] = float(mp_flux.flux(ydeg=l_max)(b, r, y=y))
        calc[i] = light_curve(y)

    for n in range(expect.size):
        # For logging/debugging purposes, work out the case id
        l = np.floor(np.sqrt(n)).astype(int)  # noqa
        m = n - l**2 - l
        mu = l - m
        nu = l + m
        if mu == 1 and l == 1:
            case = 1
        elif mu % 2 == 0 and (mu // 2) % 2 == 0:
            case = 2
        elif mu == 1 and l % 2 == 0:
            case = 3
        elif mu == 1:
            case = 4
        elif (mu - 1) % 2 == 0 and ((mu - 1) // 2) % 2 == 0:
            case = 5
        else:
            case = 0

        assert_allclose(
            calc[n],
            expect[n],
            err_msg=f"n={n}, l={l}, m={m}, mu={mu}, nu={nu}, case={case}",
            atol=1e-6,
        )


@pytest.fixture(
    params=[
        {
            "central": keplerian.Central(
                mass=1.3,
                radius=1.1,
            ),
            "bodies": [{"radius": 0.5, "mass": 0.1, "period": 1.5}],
        },
        {
            "central": keplerian.Central(
                mass=1.3,
                radius=1.1,
            ),
            "central_surface": Surface(
                y=Ylm.from_dense(
                    [1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03, 0.04, 0.4, 0.2, 0.1]
                ),
                inc=0.9,
                obl=0.3,
                period=1.2,
                u=(0.1, 0.1),
            ),
            "bodies": [
                {
                    "radius": 0.5,
                    "mass": 0.1,
                    "period": 1.5,
                    "surface": Surface(
                        y=Ylm.from_dense([1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03]),
                        inc=-0.3,
                        period=0.8,
                        u=(0.2, 0.3),
                    ),
                }
            ],
        },
        {
            "central": keplerian.Central(
                mass=1.3,
                radius=1.1,
            ),
            "central_surface": Surface(
                y=Ylm.from_dense(
                    [1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03, 0.04, 0.4, 0.2, 0.1]
                ),
                inc=0.9,
                obl=0.3,
                period=1.2,
                u=(0.1, 0.1),
            ),
            "bodies": [
                {
                    "radius": 0.5,
                    "mass": 0.1,
                    "period": 1.5,
                },
                {
                    "radius": 0.2,
                    "mass": 0.3,
                    "period": 0.4,
                },
            ],
        },
        {
            "central": keplerian.Central(
                mass=1.3,
                radius=1.1,
            ),
            "central_surface": Surface(
                y=Ylm.from_dense(np.hstack([1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03])),
                period=1.2,
                u=(0.1, 0.1),
            ),
            "bodies": [
                {
                    "radius": 0.5,
                    "mass": 0.1,
                    "period": 1.5,
                }
            ],
        },
        {
            "central": keplerian.Central(
                mass=0.3,
                radius=0.1,
            ),
            "bodies": [
                {
                    "radius": 1.5,
                    "mass": 1.1,
                    "period": 1.5,
                    "surface": Surface(
                        y=Ylm.from_dense([1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03]),
                        period=1.2,
                        u=(0.1, 0.1),
                    ),
                }
            ],
        },
    ]
)
def keplerian_system(request):
    system = SurfaceSystem(
        request.param["central"],
        central_surface=request.param.get("central_surface", None),
    )
    for body in request.param["bodies"]:
        system = system.add_body(**body)

    return system


def test_compare_starry_system(keplerian_system):
    starry = pytest.importorskip("starry")
    starry.config.lazy = False

    def jaxoplanet2starry(body, surface_map=None):
        cls = (
            starry.Primary if isinstance(body, keplerian.Central) else starry.Secondary
        )
        if surface_map is None or surface_map.period is None:
            prot = 1e15
        else:
            prot = surface_map.period

        if surface_map is None:
            map_kwargs = dict(ydeg=0, rv=False, reflected=False)
        else:
            map_kwargs = dict(
                ydeg=surface_map.ydeg,
                udeg=surface_map.udeg,
                inc=np.rad2deg(surface_map.inc),
                obl=np.rad2deg(surface_map.obl),
                amp=surface_map.amplitude,
            )

        body_kwargs = dict(
            r=body.radius.magnitude,
            m=body.mass.magnitude,
            prot=prot,
        )

        if isinstance(body, keplerian.OrbitalBody):
            body_kwargs["porb"] = body.period.magnitude
            map_kwargs["amp"] = surface_map.amplitude if surface_map else 0.0

        starry_body = cls(starry.Map(**map_kwargs), **body_kwargs)

        if surface_map and surface_map.u:
            starry_body.map[1:] = surface_map.u
        if surface_map and surface_map.deg > 0:
            starry_body.map[1:, :] = np.asarray(surface_map.y.todense())[1:]

        return starry_body

    time = np.linspace(-1.5, 1.0, 300)

    # jaxoplanet
    jaxoplanet_flux = light_curve(keplerian_system)(time)

    # starry
    primary = jaxoplanet2starry(
        keplerian_system.central, keplerian_system.central_surface
    )
    secondaries = [
        jaxoplanet2starry(body, surface_map)
        for body, surface_map in zip(
            keplerian_system.bodies, keplerian_system.body_surfaces, strict=False
        )
    ]

    starry_system = starry.System(primary, *secondaries)
    starry_flux = starry_system.flux(time, total=False)

    assert_allclose(jaxoplanet_flux.T, np.array(starry_flux))

    # # compare summed jaxponet fluxes with the corresponding starry fluxes
    # if body_map is None:
    #     assert_allclose(jax.numpy.squeeze(jaxoplanet_flux), starry_flux[0])
    # elif central_map is None:
    #     assert_allclose(jax.numpy.squeeze(jaxoplanet_flux), starry_flux[1])
    # else:
    #     assert_allclose(jax.numpy.squeeze(jaxoplanet_flux),
    # np.sum(starry_flux, axis=0))

    # # compare individual jaxpolanet fluxes with the corresponding starry fluxes
    # for jaxoplanet_flux_item, starry_flux_item in zip(jaxoplanet_flux, starry_flux):
    #     if jaxoplanet_flux_item is not None:
    #         assert_allclose(jax.numpy.squeeze(jaxoplanet_flux_item), starry_flux_item)


def test_map_light_curves_none_occultor():
    surface = Surface(
        y=Ylm.from_dense(np.hstack([1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03])),
        period=1.2,
        u=(0.1, 0.1),
    )

    expected = surface_light_curve(surface, theta=0.5)
    calc = surface_light_curve(surface, 0.0, 2.0, 2.0, 2.0, theta=0.5)

    assert_allclose(calc, expected)


@pytest.mark.parametrize("deg", [2, 5, 10])
def test_compare_starry_rot(deg):
    starry = pytest.importorskip("starry")
    starry.config.lazy = False

    # map
    inc = np.pi / 2
    np.random.seed(deg)
    y = Ylm.from_dense(np.random.randn((deg + 1) ** 2), normalize=True)
    map = Surface(y=y, inc=inc)

    # phase
    theta = np.linspace(0, 2 * np.pi, 200)

    # starry
    ms = starry.Map(ydeg=deg, inc=np.rad2deg(inc))
    ms[:, :] = y.todense()
    expected = ms.flux(theta=np.rad2deg(theta))
    calc = jax.vmap(surface_light_curve, in_axes=(None, None, None, None, None, 0))(
        map, None, None, None, None, theta
    )

    assert_allclose(calc, expected)


def test_EB():

    from jaxoplanet.units import unit_registry as ureg

    params = {
        "primary_mass": 2.0,
        "secondary_mass": 1.0,
        "primary_radius": 2.0,
        "secondary_radius": 1.0,
        "inclination": np.pi / 2,
        "period": 1.0,
        "t0": 0.0,
        "s": 0.35,
    }

    primary = keplerian.Central(
        radius=params["primary_radius"] * ureg.R_sun,
        mass=params["primary_mass"] * ureg.M_sun,
    )

    primary_amplitude = 1.1
    primary_surface = Surface(amplitude=primary_amplitude, normalize=False)

    secondary = keplerian.Body(
        radius=params["secondary_radius"] * ureg.R_sun,
        mass=params["secondary_mass"] * ureg.M_sun,
        period=params["period"] * ureg.day,
        time_transit=params["t0"] * ureg.day,
        inclination=params["inclination"] * ureg.rad,
    )

    secondary_amplitude = params["s"]
    secondary_surface = Surface(amplitude=secondary_amplitude, normalize=False)

    system = SurfaceSystem(primary, primary_surface).add_body(
        secondary, secondary_surface
    )

    def flux_function(time):
        light_curve(system)(time)

    # no occultation
    assert_allclose(
        flux_function(params["period"] / 4), [primary_amplitude, secondary_amplitude]
    )

    # primary occultation
    assert_allclose(flux_function(params["period"] / 2), [primary_amplitude, 0.0])

    # secondary occultation
    depth = (params["secondary_radius"] / params["primary_radius"]) ** 2
    assert_allclose(
        flux_function(0.0), [primary_amplitude * (1 - depth), secondary_amplitude]
    )


def test_compare_limb_dark_light_curve():
    time = np.linspace(-0.1, 0.1, 200)

    params = {
        "stellar_mass": 0.3,
        "stellar_radius": 0.3,
        "planet_radius": 0.1,
        "u": (0.1, 0.1),
        "planet_period": 15.0,
    }

    system = keplerian.System(
        keplerian.Central(mass=params["stellar_mass"], radius=params["stellar_radius"]),
    )

    system = system.add_body(
        radius=params["planet_radius"], period=params["planet_period"]
    )

    expected = limb_dark_light_curve(system, params["u"])(time)[:, 0] + 1.0

    surface_system = SurfaceSystem(
        keplerian.Central(mass=params["stellar_mass"], radius=params["stellar_radius"]),
        Surface(u=params["u"]),
    )
    surface_system = surface_system.add_body(
        radius=params["planet_radius"], period=params["planet_period"]
    )

    calc = light_curve(surface_system)(time)[:, 0]

    assert_allclose(calc, expected)


def test_EB_interchanged():

    params = {
        "primary_mass": 2.0,
        "secondary_mass": 1.0,
        "primary_radius": 2.0,
        "secondary_radius": 1.0,
        "inclination": np.pi / 2,
        "primary_amplitude": 1.1,
        "secondary_amplitude": 0.35,
        "period": 1.0,
        "t0": 0.0,
        "s": 0.35,
    }

    system_1 = SurfaceSystem(
        keplerian.Central(
            radius=params["primary_radius"] * ureg.R_sun,
            mass=params["primary_mass"] * ureg.M_sun,
        ),
        Surface(amplitude=params["primary_amplitude"], normalize=False),
    ).add_body(
        keplerian.Body(
            radius=params["secondary_radius"] * ureg.R_sun,
            mass=params["secondary_mass"] * ureg.M_sun,
            period=params["period"] * ureg.day,
            time_transit=params["t0"] * ureg.day,
            inclination=params["inclination"] * ureg.rad,
        ),
        Surface(amplitude=params["secondary_amplitude"], normalize=False),
    )

    system_2 = SurfaceSystem(
        keplerian.Central(
            radius=params["secondary_radius"] * ureg.R_sun,
            mass=params["secondary_mass"] * ureg.M_sun,
        ),
        Surface(amplitude=params["secondary_amplitude"], normalize=False),
    ).add_body(
        keplerian.Body(
            radius=params["primary_radius"] * ureg.R_sun,
            mass=params["primary_mass"] * ureg.M_sun,
            period=params["period"] * ureg.day,
            time_transit=params["t0"] * ureg.day,
            inclination=params["inclination"] * ureg.rad,
        ),
        Surface(amplitude=params["primary_amplitude"], normalize=False),
    )

    # switching primary and secondary
    time = np.linspace(0, params["period"], 200)
    flux_ordered = light_curve(system_1)(time).sum(1)
    flux_reversed = light_curve(system_2)(time + params["period"] / 2).sum(1)

    # for some reason the assert_allclose wasn't catching error here
    assert_allclose(flux_ordered, flux_reversed)


@pytest.mark.parametrize("order", [20, 100, 500])
def test_light_curves_orders(order):

    central = keplerian.Central(
        radius=1.0 * ureg.R_sun,
        mass=1.0 * ureg.M_sun,
    )

    np.random.seed(42)
    y = Ylm.from_dense([1.00, *np.random.normal(0.0, 2e-2, size=15)])
    surface = Surface(inc=1.0, obl=0.2, period=27.0, u=(0.1, 0.1), y=y)

    system = SurfaceSystem(central, surface).add_body(
        semimajor=40.0 * ureg.au,
        radius=20.0 * ureg.R_earth,
        mass=1.0 * ureg.M_earth,
        impact_param=0.2,
    )

    _ = light_curve(system, order=order)(0.0)


@pytest.mark.parametrize("deg", [2, 5, 10])
def test_compare_y_from_u(deg):
    """In this test we convert the limb darkening coefficients to spherical harmonic
    coefficients and compare the light curve providing one or the other.
    """

    u = np.ones(deg)
    r = 0.1
    b = np.linspace(0, 1 + r, 20)
    surface_u = Surface(u=u)
    surface_y = Surface(y=Ylm.from_limb_darkening(u))

    light_curve_function = jax.vmap(
        lambda surface, b: surface_light_curve(surface, r, y=b), (None, 0)
    )

    flux_u = light_curve_function(surface_u, b)
    flux_y = light_curve_function(surface_y, b)

    assert_allclose(flux_u, flux_y)


@pytest.mark.parametrize(
    "params",
    [
        {
            "r": [1.0, 0.2],  # radii
            "u": [(0.1, 0.2), ()],  # limb darkening
            "a": [1.0, 0.2],  # amplitudes
        },
        {
            "r": [0.2, 1.5],  # radii
            "u": [(0.1, 0.2), ()],  # limb darkening
            "a": [0.1, 0.2],  # amplitudes
        },
        {
            "r": [1.0, 0.1, 0.5],  # radii
            "u": [(0.1, 0.2), (), (0.3, 0.4)],  # limb darkening
            "a": [1.0, 0.2, 0.5],  # amplitudes
        },
    ],
)
def test_emission_light_curve(params):

    central = keplerian.Central(radius=params["r"][0], mass=0.5)
    system = SurfaceSystem(central, Surface(u=params["u"][0], amplitude=params["a"][0]))

    for r, u, a in zip(params["r"][1:], params["u"][1:], params["a"][1:], strict=False):
        system = system.add_body(
            radius=r, surface=Surface(u=u, amplitude=a), period=1.0
        )

    period = system.bodies[0].period.magnitude
    time = np.linspace(-0.5 * period, 0.5 * period, 1000)

    expected = light_curve(system)(time)
    calc = emission_light_curve(system)(time)

    assert_allclose(calc, expected)



import jax
import numpy as np
import pytest

from jaxoplanet.starry import Surface
from jaxoplanet.starry.ylm import ylm_spot
from jaxoplanet.test_utils import assert_allclose


@pytest.mark.parametrize("ydeg", [5, 10])
def test_spot_intensity(ydeg, radius=0.5, contrast=0.1):
    starry = pytest.importorskip("starry")
    starry.config.lazy = False

    map = starry.Map(ydeg=ydeg)
    map.spot(contrast=contrast, radius=np.rad2deg(radius), lat=0.0, lon=0.0)

    lon = np.linspace(-np.pi, np.pi, 500)
    exp = map.intensity(0, np.rad2deg(lon))

    jax_map = Surface(y=ylm_spot(ydeg)(contrast, radius, 0.0, 0.0))
    calc = jax.vmap(jax_map.intensity, in_axes=(None, 0))(0.0, lon)

    assert_allclose(calc, exp)



import jax
import jax.experimental
import jax.numpy as jnp
import jpu.numpy as jnpu
import numpy as np
import pytest

from jaxoplanet.orbits.keplerian import Body, Central, System
from jaxoplanet.test_utils import (
    assert_allclose,
    assert_quantity_allclose,
    assert_quantity_pytree_allclose,
)
from jaxoplanet.units import unit_registry as ureg


@pytest.fixture(
    params=[
        {
            "central": Central(mass=1.3, radius=1.1),
            "bodies": [
                Body(
                    mass=0.1,
                    time_transit=0.1,
                    period=12.5,
                    inclination=0.3,
                )
            ],
        },
        {
            "central": Central(mass=1.3, radius=1.1),
            "bodies": [
                Body(
                    mass=0.1,
                    time_transit=0.1,
                    period=12.5,
                    inclination=0.3,
                    eccentricity=0.3,
                    omega_peri=-1.5,
                    asc_node=0.3,
                )
            ],
        },
    ]
)
def system(request):
    return System(**request.param)


@pytest.fixture
def time():
    return jnp.linspace(-50.0, 50.0, 500) * ureg.day


def test_keplerian_central_density():
    star = Central()
    assert_quantity_allclose(
        star.density, 1.4 * ureg.g / ureg.cm**3, atol=0.01, convert=True
    )


def test_keplerian_central_from_orbit():
    sun = Central.from_orbital_properties(
        period=1.0 * ureg.yr,
        semimajor=1.0 * ureg.au,
        radius=1.0 * ureg.R_sun,
        body_mass=1.0 * ureg.M_earth,
    )
    # add large tolerance to account for lack of precision in ureg.yr
    assert_quantity_allclose(sun.mass, 1.0 * ureg.M_sun, atol=5e-5)


def test_keplerian_body_keplers_law():
    orbit = System().add_body(semimajor=1.0 * ureg.au)
    assert_quantity_allclose(
        orbit.bodies[0].period, 1.0 * ureg.year, atol=0.01, convert=True
    )

    orbit = System().add_body(period=1.0 * ureg.year)
    assert_quantity_allclose(
        orbit.bodies[0].semimajor, 1.0 * ureg.au, atol=0.01, convert=True
    )


@pytest.mark.parametrize("prefix", ["", "central_", "relative_"])
def test_keplerian_body_velocity(time, system, prefix):
    body = system.bodies[0]
    v = getattr(body, f"{prefix}velocity")(time)
    for i, v_ in enumerate(v):
        pos_func = getattr(body, f"{prefix}position")
        assert_allclose(
            v_.magnitude,
            jax.vmap(jax.grad(lambda t: pos_func(t)[i].magnitude))(time).magnitude,
        )


def test_keplerian_body_radial_velocity(time, system):
    body = system.bodies[0]
    computed = body.radial_velocity(time)
    assert computed.units == ureg.R_sun / ureg.d
    computed.to(ureg.m / ureg.s)

    computed = body.radial_velocity(time, semiamplitude=1.0)
    assert not hasattr(computed, "_magnitude")
    assert not hasattr(computed, "_units")


def test_keplerian_body_impact_parameter(system):
    body = system.bodies[0]
    x, y, z = body.relative_position(body.time_transit)
    assert_quantity_allclose(
        (jnpu.sqrt(x**2 + y**2) / system.central.radius),
        body.impact_param,
    )
    assert jnpu.all(z > 0)


def test_keplerian_body_coordinates_match_batman(time, system):
    _rsky = pytest.importorskip("batman._rsky")
    body = system.bodies[0]
    with jax.experimental.enable_x64(True):
        r_batman = _rsky._rsky(
            np.array(time.magnitude, dtype=np.float64),
            float(body.time_transit.magnitude),
            float(body.period.magnitude),
            float(body.semimajor.magnitude),
            float(body.inclination.magnitude),
            (float(body.eccentricity.magnitude) if body.eccentricity else 0.0),
            (float(body.omega_peri.magnitude) if body.omega_peri else 0.0),
            1,
            1,
        )
        m = r_batman < 100.0
        assert m.sum() > 0

        x, y, z = body.relative_position(time)
        r = jnpu.sqrt(x**2 + y**2)

        # Make sure that the in-transit impact parameter matches batman
        assert_allclose(r_batman[m], r.magnitude[m])

        # In-transit should correspond to positive z in our parameterization
        assert np.all(z.magnitude[m] > 0)

        # Therefore, when batman doesn't see a transit we shouldn't be transiting
        no_transit = z.magnitude[~m] < 0
        no_transit |= r.magnitude[~m] > 2
        assert np.all(no_transit)


def test_keplerian_body_positions_small_star(time):
    _rsky = pytest.importorskip("batman._rsky")
    with jax.experimental.enable_x64(True):
        body = (
            System(
                Central(radius=0.189, mass=0.151),
            )
            .add_body(
                period=0.4626413,
                time_transit=0.2,
                impact_param=0.5,
                eccentricity=0.1,
                omega_peri=0.1,
            )
            .bodies[0]
        )

        r_batman = _rsky._rsky(
            np.array(time.magnitude, dtype=np.float64),
            float(body.time_transit.magnitude),
            float(body.period.magnitude),
            float(body.semimajor.magnitude),
            float(body.inclination.magnitude),
            (float(body.eccentricity.magnitude) if body.eccentricity else 0.0),
            (float(body.omega_peri.magnitude) if body.omega_peri else 0.0),
            1,
            1,
        )
        m = r_batman < 100.0
        assert m.sum() > 0

        x, y, _ = body.relative_position(time)
        r = jnpu.sqrt(x**2 + y**2)
        assert_allclose(r_batman[m], r[m].magnitude)


def test_keplerian_system_stack_construction():
    sys = System().add_body(period=0.1).add_body(period=0.2)
    assert sys._body_stack.stack is not None
    assert_quantity_allclose(
        sys._body_stack.stack.period, jnp.array([0.1, 0.2]) * ureg.day
    )

    sys = (
        System()
        .add_body(period=0.1)
        .add_body(period=0.2, eccentricity=0.1, omega_peri=-1.5)
    )
    assert sys._body_stack.stack is None


def test_keplerian_system_radial_velocity():
    sys1 = System().add_body(period=0.1).add_body(period=0.2)
    sys2 = (
        System()
        .add_body(period=0.1)
        .add_body(period=0.2, eccentricity=0.0, omega_peri=0.0)
    )
    assert sys1._body_stack.stack is not None
    assert sys2._body_stack.stack is None
    with pytest.raises(ValueError):
        sys1._body_stack.stack.radial_velocity(0.0)

    for t in [0.0, jnp.linspace(0, 1, 5)]:
        assert_quantity_allclose(
            sys1.radial_velocity(t),
            sys2.radial_velocity(t),
            # TODO(dfm): I'm not sure why we need to loosen the tolerance here,
            # but the ecc=0 model doesn't give the same results as the ecc=None
            # model otherwise.
            atol={jnp.float32: 5e-6, jnp.float64: 1e-12},
        )


def test_keplerian_system_position():
    sys1 = System().add_body(period=0.1).add_body(period=0.2)
    sys2 = (
        System()
        .add_body(period=0.1)
        .add_body(period=0.2, eccentricity=0.0, omega_peri=0.0)
    )
    for t in [0.0, jnp.linspace(0, 1, 5)]:
        x1, y1, z1 = sys1.position(t)
        x2, y2, z2 = sys2.position(t)
        assert_quantity_allclose(x1, x2)
        assert_quantity_allclose(y1, y2)
        assert_quantity_allclose(z1, z2)


def body_vmap_func1(body, x):
    return body.radius * x


def body_vmap_func2(body, x):
    return body.radius * x["x"]


def body_vmap_func3(body, x):
    return {"y": body.radius * x}


def body_vmap_func4(body, x):
    del body
    return {"y": x}


@pytest.mark.parametrize(
    "func, in_axes, out_axes, args",
    [
        (body_vmap_func1, None, 0, (jnp.linspace(0.0, 2.0, 3),)),
        (body_vmap_func1, 0, 0, (jnp.linspace(0.0, 2.0, 6).reshape(2, 3),)),
        (body_vmap_func1, (0,), 0, (jnp.linspace(0.0, 2.0, 6).reshape(2, 3),)),
        (body_vmap_func1, (1,), 0, (jnp.linspace(0.0, 2.0, 6).reshape(3, 2),)),
        (body_vmap_func2, None, 0, ({"x": jnp.linspace(0.0, 2.0, 3)},)),
        (body_vmap_func2, {"x": None}, 0, ({"x": jnp.linspace(0.0, 2.0, 3)},)),
        (
            body_vmap_func2,
            {"x": 0},
            0,
            ({"x": jnp.linspace(0.0, 2.0, 6).reshape(2, 3)},),
        ),
        (body_vmap_func3, None, 0, (jnp.linspace(0.0, 2.0, 3),)),
        (body_vmap_func3, None, 1, (jnp.linspace(0.0, 2.0, 3),)),
        (body_vmap_func3, None, {"y": 1}, (jnp.linspace(0.0, 2.0, 3),)),
        (body_vmap_func4, None, None, (jnp.linspace(0.0, 2.0, 3),)),
    ],
)
def test_body_vmap(func, in_axes, out_axes, args):
    central = Central()
    vmap_sys = (
        System(central)
        .add_body(radius=0.5, period=1.0)
        .add_body(radius=0.8, period=1.0)
    )
    no_vmap_sys = (
        System(central)
        .add_body(radius=0.5, period=1.0, eccentricity=0.1, omega_peri=0.1)
        .add_body(radius=0.8, period=1.0)
    )

    result1 = vmap_sys.body_vmap(func, in_axes=in_axes, out_axes=out_axes)(*args)
    result2 = no_vmap_sys.body_vmap(func, in_axes=in_axes, out_axes=out_axes)(*args)
    assert_quantity_pytree_allclose(result1, result2)



from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
from scipy.special import roots_legendre

from jaxoplanet.core.limb_dark import kite_area
from jaxoplanet.types import Array
from jaxoplanet.utils import zero_safe_sqrt


def solution_vector(l_max: int, order: int = 20) -> Callable[[Array, Array], Array]:
    n_max = l_max**2 + 2 * l_max + 1

    @jax.jit
    @partial(jnp.vectorize, signature=f"(),()->({n_max})")
    def impl(b: Array, r: Array) -> Array:
        b = jnp.abs(b)
        r = jnp.abs(r)
        kappa0, kappa1 = kappas(b, r)
        P = p_integral(order, l_max, b, r, kappa0)
        Q = q_integral(l_max, 0.5 * jnp.pi - kappa1)
        return Q - P

    return impl


def kappas(b: Array, r: Array) -> tuple[Array, Array]:
    b2 = jnp.square(b)
    factor = (r - 1) * (r + 1)
    b_cond = jnp.logical_and(jnp.greater(b, jnp.abs(1 - r)), jnp.less(b, 1 + r))
    b_ = jnp.where(b_cond, b, 1)
    area = jnp.where(b_cond, kite_area(r, b_, 1), 0)
    return jnp.arctan2(area, b2 + factor), jnp.arctan2(area, b2 - factor)


def q_integral(l_max: int, lam: Array) -> Array:
    zero = jnp.zeros_like(lam)
    c = jnp.cos(lam)
    s = jnp.sin(lam)
    h = {
        (0, 0): 2 * lam + jnp.pi,
        (0, 1): -2 * c,
    }

    def get(u: int, v: int) -> Array:
        if (u, v) in h:
            return h[(u, v)]
        if u >= 2:
            comp = 2 * c ** (u - 1) * s ** (v + 1) + (u - 1) * get(u - 2, v)
        else:
            assert v >= 2
            comp = -2 * c ** (u + 1) * s ** (v - 1) + (v - 1) * get(u, v - 2)
        comp /= u + v
        h[(u, v)] = comp
        return comp

    U = []
    for l in range(l_max + 1):  # noqa
        for m in range(-l, l + 1):
            if l == 1 and m == 0:
                U.append((np.pi + 2 * lam) / 3)
                continue
            mu = l - m
            nu = l + m
            if (mu % 2) == 0 and (mu // 2) % 2 == 0:
                u = mu // 2 + 2
                v = nu // 2
                assert u % 2 == 0
                U.append(get(u, v))
            else:
                U.append(zero)

    return jnp.stack(U)


def p_integral(order: int, l_max: int, b: Array, r: Array, kappa0: Array) -> Array:
    """Numerical integration of the P integral using the Gauss-Legendre quadrature.

    As described in Equation D32 of Luger et al. (2019), there are 6 cases to consider.
    Empirically, we notice that the numerical integration of the first case (mu/2 even)
    is precise at very low order. Hence ``low_order=30``is used for the first case. For
    the other cases, we use the order specified by the user, renamed in the function
    ``high_order``. We also note that outside the linear limb-darkening case (i.e.
    (l,m)=(1, 0), or n=2) the integrand is symmetrical in phi, so we can evaluate the
    integral over half the range and multiply by 2.


    Parameters
    ----------
    order : int
        The order of the Gauss-Legendre quadrature.
    l_max : int
        The maximum degree of the spherical harmonic expansion.
    b : Array
        Impact parameter.
    r : Array
        Occultor radius.
    kappa0 : Array
        k0 angle.

    Returns
    -------
    Array
        The integral of the P function over the occultor surface.
    """
    b2 = jnp.square(b)
    r2 = jnp.square(r)

    # This is a hack for when r -> 0 or b -> 0, so k2 -> inf
    factor = 4 * b * r
    k2_cond = jnp.less(factor, 10 * jnp.finfo(factor.dtype).eps)
    factor = jnp.where(k2_cond, 1, factor)
    k2 = jnp.maximum(0, (1 - r2 - b2 + 2 * b * r) / factor)
    # And for when r -> 0
    r_cond = jnp.less(r, 10 * jnp.finfo(r.dtype).eps)
    delta = (b - r) / (2 * jnp.where(r_cond, 1, r))
    rng = 0.25 * kappa0

    # low order variables
    low_order = np.min([order, 20])
    roots, low_weights = roots_legendre(low_order)
    phi = rng * (roots + 1)
    low_s2 = jnp.square(jnp.sin(phi))
    low_a1 = low_s2 - jnp.square(low_s2)
    low_a2 = jnp.where(r_cond, 0, delta + low_s2)

    # high order variables
    high_order = order
    high_roots, high_weights = roots_legendre(high_order)
    phi = rng * (high_roots + 1)
    high_s2 = jnp.square(jnp.sin(phi))
    high_f0 = jnp.maximum(0, jnp.where(k2_cond, 1 - r2, factor * (k2 - high_s2))) ** 1.5
    high_a1 = high_s2 - jnp.square(high_s2)
    high_a2 = jnp.where(r_cond, 0, delta + high_s2)
    high_a4 = 1 - 2 * high_s2

    low_indices = []
    low_integrand = []

    high_indices = []
    high_integrand = []

    n = 0

    for l in range(l_max + 1):  # noqa
        high_fa3 = (2 * r) ** (l - 1) * high_f0
        for m in range(-l, l + 1):
            mu = l - m
            nu = l + m

            if mu == 1 and l == 1:
                phi = 2 * rng * high_roots
                c = jnp.cos(phi + 0.5 * kappa0)
                omz2 = r2 + b2 - 2 * b * r * c
                cond = jnp.less(omz2, 10 * jnp.finfo(omz2.dtype).eps)
                omz2 = jnp.where(cond, 1, omz2)
                z2 = jnp.maximum(0, 1 - omz2)
                result = (
                    2 * r * (r - b * c) * (1 - z2 * zero_safe_sqrt(z2)) / (3 * omz2)
                )
                high_integrand.append(jnp.where(cond, 0, 2 * result))
                high_indices.append(n)

            elif mu % 2 == 0 and (mu // 2) % 2 == 0:
                f = (
                    2
                    * (2 * r) ** (l + 2)
                    * low_a1 ** (0.25 * (mu + 4))
                    * low_a2 ** (0.5 * nu)
                )
                low_integrand.append(2 * f)
                low_indices.append(n)

            elif mu == 1 and l % 2 == 0:
                f = high_fa3 * high_a1 ** (l // 2 - 1) * high_a4
                high_integrand.append(2 * f)
                high_indices.append(n)

            elif mu == 1:
                f = high_fa3 * high_a1 ** ((l - 3) // 2) * high_a2 * high_a4
                high_integrand.append(2 * f)
                high_indices.append(n)

            elif (mu - 1) % 2 == 0 and ((mu - 1) // 2) % 2 == 0:
                f = (
                    2
                    * high_fa3
                    * high_a1 ** ((mu - 1) // 4)
                    * high_a2 ** (0.5 * (nu - 1))
                )
                high_integrand.append(2 * f)
                high_indices.append(n)

            else:
                n += 1
                continue

            n += 1

    low_indices = np.stack(low_indices)
    high_indices = np.stack(high_indices)

    low_P0 = rng * jnp.sum(jnp.stack(low_integrand) * low_weights, axis=1)
    high_P0 = rng * jnp.sum(jnp.stack(high_integrand) * high_weights, axis=1)

    P = jnp.zeros(l_max**2 + 2 * l_max + 1)
    P = P.at[low_indices].set(low_P0)
    P = P.at[high_indices].set(high_P0)

    return P


def rT(lmax: int) -> Array:
    rt = [0.0 for _ in range((lmax + 1) * (lmax + 1))]
    amp0 = jnp.pi
    lfac1 = 1.0
    lfac2 = 2.0 / 3.0
    for ell in range(0, lmax + 1, 4):
        amp = amp0
        for m in range(0, ell + 1, 4):
            mu = ell - m
            nu = ell + m
            rt[ell * ell + ell + m] = amp * lfac1
            rt[ell * ell + ell - m] = amp * lfac1
            if ell < lmax:
                rt[(ell + 1) * (ell + 1) + ell + m + 1] = amp * lfac2
                rt[(ell + 1) * (ell + 1) + ell - m + 1] = amp * lfac2
            amp *= (nu + 2.0) / (mu - 2.0)
        lfac1 /= (ell / 2 + 2) * (ell / 2 + 3)
        lfac2 /= (ell / 2 + 2.5) * (ell / 2 + 3.5)
        amp0 *= 0.0625 * (ell + 2) * (ell + 2)

    amp0 = 0.5 * jnp.pi
    lfac1 = 0.5
    lfac2 = 4.0 / 15.0
    for ell in range(2, lmax + 1, 4):
        amp = amp0
        for m in range(2, ell + 1, 4):
            mu = ell - m
            nu = ell + m
            rt[ell * ell + ell + m] = amp * lfac1
            rt[ell * ell + ell - m] = amp * lfac1
            if ell < lmax:
                rt[(ell + 1) * (ell + 1) + ell + m + 1] = amp * lfac2
                rt[(ell + 1) * (ell + 1) + ell - m + 1] = amp * lfac2
            amp *= (nu + 2.0) / (mu - 2.0)
        lfac1 /= (ell / 2 + 2) * (ell / 2 + 3)
        lfac2 /= (ell / 2 + 2.5) * (ell / 2 + 3.5)
        amp0 *= 0.0625 * ell * (ell + 4)
    return np.array(rt)



import numpy as np

from jaxoplanet.core.limb_dark import light_curve as ld_light_curve
from jaxoplanet.starry import Surface, Ylm
from jaxoplanet.starry.light_curves import surface_light_curve
from jaxoplanet.starry.multiprecision import flux, mp


def starry_light_curve(
    ydeg: int, udeg: int, r: float, order: int, b: float | None = None
) -> float:
    """Estimate the absolute precision of a starry light curve.

    Args:
        ydeg (int): Degree of the spherical harmonics map.
        udeg (int): Order of the limb darkening polynomial.
        r (float): Radius ratio.
        order (int): Numerical integration order.
        b (float, optional): Impact parameter. Defaults to None.

    Returns:
        float: Precision of the jaxoplanet light curve model.
    """
    if b is None:
        b = 1 - r if r < 1.0 else r
    u = [1.0] * udeg
    y = np.ones((ydeg + 1) ** 2)
    surface = Surface(y=Ylm.from_dense(y), u=u, normalize=False)
    expected = flux.flux(ydeg=ydeg, udeg=udeg)(b, r, u=u, y=y)
    calc = surface_light_curve(surface, y=b, z=1.0, r=r, order=order)
    return float(np.abs(float(mp.mpf(float(calc)) - expected)))


def limb_dark_light_curve(
    udeg: int, r: float, order: int, b: float | None = None
) -> float:
    """Estimate the absolute precision of a limb darkened star light curve.

    Args:
        udeg (int): Order of the limb darkening polynomial.
        r (float): Radius ratio.
        order (int): Numerical integration order.
        b (float | None, optional): Impact parameter. Defaults to None.

    Returns:
        float: Precision of the jaxoplanet light curve model.
    """
    if b is None:
        b = 1 - r if r < 1.0 else r
    u = [1.0] * udeg
    expected = flux.flux(udeg=udeg)(b, r, u=u)
    calc = 1 + ld_light_curve(u, b, r, order=order)
    return np.abs(float(mp.mpf(float(calc)) - expected))



"""A module providing decorators to transform light curve functions
"""

__all__ = ["integrate", "interpolate"]

from functools import wraps
from typing import Any

import jax
import jax.numpy as jnp
import jpu.numpy as jnpu
from jpu.core import is_quantity

from jaxoplanet import units
from jaxoplanet.light_curves.types import LightCurveFunc
from jaxoplanet.light_curves.utils import vectorize
from jaxoplanet.types import Array, Quantity
from jaxoplanet.units import unit_registry as ureg

try:
    from jax.extend import linear_util as lu
except ImportError:
    from jax import linear_util as lu  # type: ignore


@units.quantity_input(exposure_time=ureg.d)
def integrate(
    func: LightCurveFunc,
    exposure_time: Quantity | None = None,
    order: int = 0,
    num_samples: int = 7,
) -> LightCurveFunc:
    """Transform a light curve function to apply exposure time integration

    This transformation applies a fixed stencil numerical integration scheme to the input
    function ``func`` to convolve the light curve with a top hat exposure time centered
    on the input time, with a full width of ``exposure_time``.

    The order of the integration scheme is set using the ``order`` parameter which must
    be ``0``, ``1``, or ``2``. The default (``0``) uses the "resampling" scheme discussed
    by `Kipping (2010) <https://arxiv.org/abs/1004.3741>`_. The higher order schemes
    ``1`` and ``2`` apply the trapezoid and Simpson's rules respectively, but won't
    necessarily provide higher accuracy results because of discontinuities at the
    contact points.

    In practice, the parameter ``num_samples`` which sets the number of function
    evaluations per integral has the most significant effect on the accuracy of this
    integral, trading off against higher computational cost.

    Args:
        func: A light curve function which takes a time ``Quantity`` as the first
            argument
        exposure_time (Quantity): The exposure time (in days, by default)
        order (int): The order of the integration scheme as discussed above
        num_samples (int): The number of function evaluations made per integral,
            controlling the accuracy of the numerics

    Returns:
        A new light curve function with the same signature as ``func``, computing the
        exposure time integrated flux
    """
    if exposure_time is None:
        return func

    if jnpu.ndim(exposure_time) != 0:
        raise ValueError(
            "The exposure time passed to 'integrate_exposure_time' has shape "
            f"{jnpu.shape(exposure_time)}, but a scalar was expected; "
            "To use exposure time integration with different exposures at different "
            "times, manually 'vmap' or 'vectorize' the function"
        )

    # Ensure 'num_samples' is an odd number
    num_samples = int(num_samples)
    num_samples += 1 - num_samples % 2
    stencil = jnp.ones(num_samples)

    # Construct exposure time integration stencil
    if order == 0:
        dt = jnp.linspace(-0.5, 0.5, 2 * num_samples + 1)[1:-1:2]
    elif order == 1:
        dt = jnp.linspace(-0.5, 0.5, num_samples)
        stencil = 2 * stencil
        stencil = stencil.at[0].set(1)
        stencil = stencil.at[-1].set(1)
    elif order == 2:
        dt = jnp.linspace(-0.5, 0.5, num_samples)
        stencil = stencil.at[1:-1:2].set(4)
        stencil = stencil.at[2:-1:2].set(2)
    else:
        raise ValueError(
            "The parameter 'order' in 'integrate_exposure_time' must be 0, 1, or 2"
        )
    dt = dt * exposure_time
    stencil /= jnp.sum(stencil)

    @wraps(func)
    @units.quantity_input(time=ureg.d)
    @vectorize
    def wrapped(time: Quantity, *args: Any, **kwargs: Any) -> Array | Quantity:
        if jnpu.ndim(time) != 0:
            raise ValueError(
                "The time passed to 'integrate_exposure_time' has shape "
                f"{jnpu.shape(time)}, but a scalar was expected; "
                "this shouldn't typically happen so please open an issue "
                "on GitHub demonstrating the problem"
            )

        f = lu.wrap_init(jax.vmap(func, in_axes=(0,) + (None,) * len(args)))
        f = apply_exposure_time_integration(f, stencil, dt)  # type: ignore
        return f.call_wrapped(time, args, kwargs)  # type: ignore

    return wrapped


@lu.transformation  # type: ignore
def apply_exposure_time_integration(stencil, dt, time, args, kwargs):
    result = yield (time + dt,) + args, kwargs
    yield jnpu.dot(stencil, result)


@units.quantity_input(period=ureg.day, time_transit=ureg.day, duration=ureg.day)
def interpolate(
    func: LightCurveFunc,
    *,
    period: Quantity,
    time_transit: Quantity,
    num_samples: int,
    duration: Quantity | None = None,
    args: tuple[Any, ...] = (),
    kwargs: dict[str, Any] | None = None,
) -> LightCurveFunc:
    """Transform a light curve function to pre-compute the model on a grid

    Sometimes it can be useful to precompute the light curve on a grid near a transit,
    and then interpolate those computations to the required phases when computing the
    full model. This can speed things up a lot when you have many transits, or a lot of
    out of transit data. This transform uses linear interpolation.

    .. note:: Unlike some other transforms, this function requires that any upstream
        ``*args`` and ``**kwargs`` be passed directly to the transform, rather than when
        calling the transformed function. This is necessary because the model is
        pre-computed when it is tranformed.

    Args:
        func: A light curve function which takes a time ``Quantity`` as the first
            argument
        period (Quantity): The period of the orbit. Used to wrap the input times into the
            domain of the pre-computed model
        time_transit (Quantity): The transit time of the orbit. Used to wrap the input
            times into the domain of the pre-computed model
        duration (Quantity): The duration centered on the transit to pre-compute. By
            default, the full period will be evaluated
        num_samples (int): The number of points in the time grid used for pre-computation
        args (tuple): Any extra positional arguments that should be passed to ``func``
        kwargs (dict): Any extra keyword arguments that should be passed to ``func``

    Returns:
        A new light curve function with the same signature as ``func``, computing the
        flux by interpolating a pre-computed model
    """

    kwargs = kwargs or {}
    if duration is None:
        duration = period
    time_grid = time_transit + duration * jnp.linspace(-0.5, 0.5, num_samples)
    flux_grid = func(time_grid, *args, **kwargs)

    if is_quantity(flux_grid):
        flux_magnitude = flux_grid.magnitude
        flux_units = flux_grid.units
    else:
        flux_magnitude = flux_grid
        flux_units = None

    @wraps(func)
    @units.quantity_input(time=ureg.d)
    @vectorize
    def wrapped(time: Quantity, *args: Any, **kwargs: Any) -> Array | Quantity:
        del args, kwargs
        time_wrapped = (
            jnpu.mod(time - time_transit + 0.5 * period, period)
            + 0.5 * period
            + time_transit
        )
        flux = jnp.interp(
            time_wrapped.magnitude,
            time_grid.magnitude,
            flux_magnitude,
            left=flux_magnitude[0],
            right=flux_magnitude[-1],
            period=period.magnitude,
        )
        if flux_units is None:
            return flux
        else:
            return flux * flux_units

    return wrapped



# mypy: ignore-errors

import warnings

import jax
import numpy as np
import pytest
from jax.test_util import check_grads

from jaxoplanet.core.limb_dark import light_curve
from jaxoplanet.test_utils import assert_allclose

exoplanet_core = pytest.importorskip("exoplanet_core")


@pytest.mark.parametrize("r", [0.01, 0.1, 1.1, 2.0])
def test_compare_exoplanet(r):
    u1 = 0.2
    u2 = 0.3
    b = np.linspace(-1 - 2 * r, 1 + 2 * r, 5001)
    expect = exoplanet_core.quad_limbdark_light_curve(u1, u2, b, r)
    calc = jax.jit(light_curve)(np.array([u1, u2]), b, r)
    assert_allclose(calc, expect)


@pytest.mark.parametrize("u", [[0.2], [0.2, 0.3], [0.2, 0.3, 0.1, 0.5, 0.02]])
@pytest.mark.parametrize("r", [0.01, 0.1, 0.5, 1.1, 2.0])
def test_edge_cases(u, r):
    u = np.array(u)
    for b in [0.0, 0.5, 1.0, r, np.abs(1 - r), 1 + r]:
        calc = jax.jit(light_curve)(u, b, r)
        assert np.isfinite(calc)
        if len(u) == 2:
            expect = exoplanet_core.quad_limbdark_light_curve(*u, b, r)
            assert_allclose(calc, expect[0])

        for n in range(3):
            g = jax.grad(light_curve, argnums=n)(u, b, r)
            assert np.all(np.isfinite(g))

    if jax.config.jax_enable_x64:  # type: ignore
        for b in [0.0, 0.5, 1.0, r, 1 + 2 * r]:
            if np.allclose(b, r) or np.allclose(np.abs(b - r), 1):
                continue
            check_grads(light_curve, (u, b, r), order=1)


@pytest.mark.parametrize("u", [[0.2], [0.2, 0.3], [0.2, 0.3, 0.1, 0.5, 0.02]])
@pytest.mark.parametrize("r", [0.01, 0.1, 0.5, 1.1, 2.0])
def test_compare_starry(u, r):
    u = np.array(u)
    starry = pytest.importorskip("starry")
    theano = pytest.importorskip("theano")
    theano.config.gcc__cxxflags += " -fexceptions"

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        m = starry.Map(udeg=len(u))
        m[1:] = u
        b_ = theano.tensor.dscalar()
        func = theano.function([b_], m.flux(xo=b_, yo=0.0, zo=1.0, ro=r) - 1)
        for b in [0.0, 0.5, 1.0, r, np.abs(1 - r), 1 + r]:
            expect = func(b)[0]
            if not np.isfinite(expect):
                continue  # hack because starry doesn't handle all edge cases properly
            calc = light_curve(u, b, r)
            assert_allclose(calc, expect)

        b = np.linspace(-1 - 2 * r, 1 + 2 * r, 5001)
        expect = m.flux(xo=b, yo=0.0, zo=1.0, ro=r).eval() - 1
        calc = light_curve(u, b, r)
        assert_allclose(calc, expect)



import jax.numpy as jnp
import pytest

from jaxoplanet.light_curves.utils import vectorize
from jaxoplanet.units import quantity_input, unit_registry as ureg


@pytest.mark.parametrize("shape", [(), (10,), (10, 3)])
def test_vectorize_scalar(shape):
    @quantity_input(time=ureg.day)
    @vectorize
    def lc(time):
        assert time.shape == ()
        return time.magnitude**2

    time = jnp.ones(shape)
    assert lc(time).shape == time.shape


@pytest.mark.parametrize("shape", [(), (10,), (10, 3)])
def test_vectorize_array(shape):
    @vectorize
    def lc(time):
        assert time.shape == ()
        return jnp.array([time, time**2, 1.0])

    time = jnp.ones(shape)
    assert lc(time).shape == time.shape + (3,)


@pytest.mark.parametrize("shape", [(), (10,), (10, 3)])
@pytest.mark.parametrize("out_units", [None, ureg.dimensionless, ureg.m])
def test_vectorize_quantity(shape, out_units):
    @quantity_input(time=ureg.day)
    @vectorize
    def lc(time, units):
        assert time.shape == ()
        y = jnp.stack([time.magnitude, time.magnitude**2, 1.0])
        if units is None:
            return y
        else:
            return y * units

    time = jnp.ones(shape)
    assert lc(time, out_units).shape == time.shape + (3,)
    assert lc(time, units=out_units).shape == time.shape + (3,)



from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp

from jaxoplanet.light_curves.utils import vectorize
from jaxoplanet.starry.orbit import SurfaceSystem
from jaxoplanet.types import Array, Quantity
from jaxoplanet.units import quantity_input, unit_registry as ureg


def system_observable(surface_observable, **kwargs):
    _surface_observable = partial(surface_observable, **kwargs)

    def observable_fun(
        system: SurfaceSystem,
    ) -> Callable[[Quantity], tuple[Array | None, Array | None]]:
        # the observable function of the central given the position and radii
        # of all other bodies
        central_bodies_observable = jax.vmap(
            _surface_observable, in_axes=(None, 0, 0, 0, 0, None)
        )

        # the observable function of all bodies combined given their position to the
        # central
        @partial(system.surface_vmap, in_axes=(0, 0, 0, 0, None))
        def compute_body_observable(surface, radius, x, y, z, time):
            if surface is None:
                return 0.0
            else:
                theta = surface.rotational_phase(time.magnitude)
                return _surface_observable(
                    surface,
                    (system.central.radius / radius).magnitude,
                    (x / radius).magnitude,
                    (y / radius).magnitude,
                    (z / radius).magnitude,
                    theta,
                )

        @quantity_input(time=ureg.day)
        @vectorize
        def obsrvable_impl(time: Quantity) -> Array:
            # a function that give the array of observables for all bodies, starting
            # with the central
            if system.central_surface is None:
                central_light_curves = jnp.array([0.0])
            else:
                theta = system.central_surface.rotational_phase(time.magnitude)
                central_radius = system.central.radius
                central_phase_curve = surface_observable(
                    system.central_surface, theta=theta
                )
                if len(system.bodies) > 0:
                    xos, yos, zos = system.relative_position(time)
                    n = len(xos.magnitude)
                    central_light_curves = central_bodies_observable(
                        system.central_surface,
                        (system.radius / central_radius).magnitude,
                        (xos / central_radius).magnitude,
                        (yos / central_radius).magnitude,
                        (zos / central_radius).magnitude,
                        theta,
                    )

                    if n > 1 and central_light_curves is not None:
                        central_light_curves = central_light_curves.sum(
                            0
                        ) - central_phase_curve * (n - 1)
                        central_light_curves = jnp.expand_dims(central_light_curves, 0)

                    body_light_curves = compute_body_observable(
                        system.radius, -xos, -yos, -zos, time
                    )

                    return jnp.hstack([central_light_curves, body_light_curves])
                else:
                    return jnp.array([central_phase_curve])

        return obsrvable_impl

    return observable_fun



# mypy: ignore-errors

import jax
import jax.numpy as jnp
import pytest
from jax.test_util import check_grads

from jaxoplanet.core import kepler
from jaxoplanet.test_utils import assert_allclose


def get_mean_and_true_anomaly(eccentricity, eccentric_anomaly):
    M = eccentric_anomaly - eccentricity * jnp.sin(eccentric_anomaly)
    f = 2 * jnp.arctan(
        jnp.sqrt((1 + eccentricity) / (1 - eccentricity))
        * jnp.tan(0.5 * eccentric_anomaly)
    )
    return M, f


def check_kepler(e, E):
    # TODO: Add check for gradients
    E = E % (2 * jnp.pi)
    M, f = get_mean_and_true_anomaly(e, E)
    sinf0, cosf0 = kepler(M, e)
    assert_allclose(sinf0, jnp.sin(f))
    assert_allclose(cosf0, jnp.cos(f))


def test_basic_low_e():
    e = jnp.linspace(0, 0.85, 500)
    E = jnp.linspace(-300, 300, 1001)
    e = e[None, :] + jnp.zeros((E.shape[0], e.shape[0]))
    E = E[:, None] + jnp.zeros_like(e)
    check_kepler(e, E)


@pytest.mark.skipif(
    not jax.config.jax_enable_x64,
    reason="Kepler solver has numerical issues at single point precision",
)
def test_basic_high_e():
    e = jnp.linspace(0.85, 1.0, 500)[:-1]
    E = jnp.linspace(-300, 300, 1001)
    e = e[None, :] + jnp.zeros((E.shape[0], e.shape[0]))
    E = E[:, None] + jnp.zeros_like(e)
    check_kepler(e, E)


@pytest.mark.skipif(
    not jax.config.jax_enable_x64,
    reason="Kepler solver has numerical issues at single point precision",
)
def test_edge_cases():
    E = jnp.array([0.0, 2 * jnp.pi, -226.2, -170.4])
    e = (1 - 1e-6) * jnp.ones_like(E)
    e = jnp.append(e[:-1], 0.9939879759519037)
    check_kepler(e, E)


def test_pi():
    e = jnp.linspace(0, 1.0, 100)[:-1]
    E = jnp.pi + jnp.zeros_like(e)
    check_kepler(e, E)


@pytest.mark.parametrize("e", [0.01, 0.1, 0.5, 0.6])
@pytest.mark.xfail(
    not jax.config.jax_enable_x64,
    reason="Gradients can be noisy at single point precision",
)
def test_grad(e):
    M = jnp.linspace(-5, 5, 100)
    check_grads(kepler, (M, e), order=1)



"""This module provides the functions needed to compute a limb darkened light curve as
described by `Agol et al. (2020) <https://arxiv.org/abs/1908.03222>`_.
"""

__all__ = ["light_curve"]

from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
from scipy.special import binom, roots_legendre

from jaxoplanet.types import Array
from jaxoplanet.utils import zero_safe_sqrt


@partial(jax.jit, static_argnames=("order",))
def light_curve(u: Array, b: Array, r: Array, *, order: int = 10):
    """Compute the light curve for arbitrary polynomial limb darkening

    See `Agol et al. (2020) <https://arxiv.org/abs/1908.03222>`_ for more technical
    details. Unlike in that paper, here we don't evaluate all the solution vector
    integrals in closed form. Instead, for all but the lowest order terms, we numerically
    integrate the relevant 1D line integral using Gauss-Legendre quadrature at fixed
    order ``order``.


    Args:
        u (Array): The coefficients of the polynomial limb darkening model
        b (Array): The center-to-center distance between the occultor and the occulted
            body
        r (Array): The radius ratio between the occultor and the occulted body
        order (int): The quadrature order to use when numerically computing the the 1D
            line integral
    """

    u = jnp.asarray(u)
    assert u.ndim == 1
    if u.shape[0] == 0:
        g = jnp.full((1,), 1.0 / jnp.pi)
    else:
        g = greens_basis_transform(u)
        g /= jnp.pi * (g[0] + g[1] / 1.5)
    s = solution_vector(len(g) - 1, order=order)(b, r)
    return s @ g - 1


def solution_vector(l_max: int, order: int = 10) -> Callable[[Array, Array], Array]:
    n_max = l_max + 1

    @partial(jnp.vectorize, signature=f"(),()->({n_max})")
    def impl(b: Array, r: Array) -> Array:
        # TODO: Are all of these conditions necessary?
        b = jnp.abs(b)
        r = jnp.abs(r)
        area, kappa0, kappa1 = kappas(b, r)

        no_occ = jnp.greater_equal(b, 1 + r)
        full_occ = jnp.less_equal(1 + b, r)
        cond = jnp.logical_or(no_occ, full_occ)
        b_: Array = jnp.where(cond, jnp.ones_like(b), b)

        b2 = jnp.square(b_)
        r2 = jnp.square(r)

        s0, s2 = s0s2(b_, r, b2, r2, area, kappa0, kappa1)
        s0: Array = jnp.where(no_occ, jnp.pi, s0)
        s0 = jnp.where(full_occ, jnp.zeros_like(s0), s0)
        s2 = jnp.where(cond, jnp.zeros_like(s2), s2)

        s = [s0[None]]
        if l_max >= 1:
            P: Array = p_integral(order, l_max, b_, r, b2, r2, kappa0)
            P = jnp.where(cond, jnp.zeros_like(P), P)
            s.append(-P[:1] - 2 * (kappa1 - jnp.pi) / 3)
            if l_max >= 2:
                s.append(s2[None])  # type: ignore
            if l_max >= 3:
                s.append(-P[1:])
        return jnp.concatenate(s, axis=0)

    return impl


def greens_basis_transform(u: Array) -> Array:
    dtype = jnp.dtype(u)
    u = jnp.concatenate((-jnp.ones(1, dtype=dtype), u))
    size = len(u)
    i = np.arange(size)
    arg = binom(i[None, :], i[:, None]) @ u
    p = (-1) ** (i + 1) * arg
    g = [jnp.zeros((), dtype=dtype) for _ in range(size + 2)]
    for n in range(size - 1, 1, -1):
        g[n] = p[n] / (n + 2) + g[n + 2]
    g[1] = p[1] + 3 * g[3]
    g[0] = p[0] + 2 * g[2]
    return jnp.stack(g[:-2])


def kappas(b: Array, r: Array) -> tuple[Array, Array, Array]:
    b2 = jnp.square(b)
    factor = (r - 1) * (r + 1)
    cond = jnp.logical_and(jnp.greater(b, jnp.abs(1 - r)), jnp.less(b, 1 + r))
    b_ = jnp.where(cond, b, jnp.ones_like(b))
    area: Array = jnp.where(cond, kite_area(r, b_, jnp.ones_like(r)), jnp.zeros_like(r))
    return area, jnp.arctan2(area, b2 + factor), jnp.arctan2(area, b2 - factor)


def s0s2(
    b: Array,
    r: Array,
    b2: Array,
    r2: Array,
    area: Array,
    kappa0: Array,
    kappa1: Array,
) -> tuple[Array, Array]:
    bpr = b + r
    onembpr2 = (1 + bpr) * (1 - bpr)
    eta2 = 0.5 * r2 * (r2 + 2 * b2)

    # Large k
    s0_lrg = jnp.pi * (1 - r2)
    s2_lrg = 2 * s0_lrg + 4 * jnp.pi * (eta2 - 0.5)

    # Small k
    Alens = kappa1 + r2 * kappa0 - area * 0.5
    s0_sml = jnp.pi - Alens
    s2_sml = 2 * s0_sml + 2 * (
        -(jnp.pi - kappa1) + 2 * eta2 * kappa0 - 0.25 * area * (1 + 5 * r2 + b2)
    )

    delta = 4 * b * r
    cond = jnp.greater(onembpr2 + delta, delta)
    return jnp.where(cond, s0_lrg, s0_sml), jnp.where(cond, s2_lrg, s2_sml)


def p_integral(
    order: int,
    l_max: int,
    b: Array,
    r: Array,
    b2: Array,
    r2: Array,
    kappa0: Array,
) -> Array:
    # This is a hack for when r -> 0 or b -> 0, so k2 -> inf
    factor = 4 * b * r
    k2_cond = jnp.less(factor, 10 * jnp.finfo(factor.dtype).eps)
    factor: Array = jnp.where(k2_cond, jnp.ones_like(factor), factor)
    k2 = jnp.maximum(jnp.zeros_like(factor), (1 - r2 - b2 + 2 * b * r) / factor)

    roots, weights = roots_legendre(order)
    rng = 0.5 * kappa0
    phi = rng * roots
    c = jnp.cos(phi + 0.5 * kappa0)

    # integrand is symmetrical so we change integration limits
    s = jnp.sin((phi + rng) / 2)
    s2 = jnp.square(s)

    arg = []
    if l_max >= 1:
        omz2 = jnp.maximum(0, r2 + b2 - 2 * b * r * c)
        z2 = 1 - omz2
        m = jnp.less(z2, 10 * jnp.finfo(omz2.dtype).eps)
        z2 = jnp.where(m, jnp.ones_like(z2), z2)
        z3 = z2 * zero_safe_sqrt(z2)
        cond = jnp.less(omz2, 10 * jnp.finfo(omz2.dtype).eps)
        omz2 = jnp.where(cond, jnp.ones_like(z2), omz2)
        result = 2 * r * (r - b * c) * (1 - z3) / (3 * omz2)
        arg.append(jnp.where(cond, jnp.zeros_like(result[None, :]), result[None, :]))
    if l_max >= 3:
        f0 = jnp.maximum(
            jnp.zeros_like(r2), jnp.where(k2_cond, 1 - r2, factor * (k2 - s2))
        )
        n = jnp.arange(3, l_max + 1)
        f = f0[None, :] ** (0.5 * n[:, None])
        f *= 2 * r * (r - b + 2 * b * s2[None, :])
        arg.append(f)

    return rng * jnp.sum(jnp.concatenate(arg, axis=0) * weights[None, :], axis=1)


def kite_area(a: Array, b: Array, c: Array) -> Array:
    def sort2(a: Array, b: Array) -> tuple[Array, Array]:
        return jnp.minimum(a, b), jnp.maximum(a, b)

    a, b = sort2(a, b)
    b, c = sort2(b, c)
    a, b = sort2(a, b)

    square_area = (a + (b + c)) * (c - (a - b)) * (c + (a - b)) * (a + (b - c))
    return zero_safe_sqrt(jnp.maximum(0, square_area))



from importlib.resources import as_file, files

import jpu

unit_registry = jpu.UnitRegistry()
with as_file(
    files("jaxoplanet.units").joinpath("astro_constants_and_units.txt")
) as path:
    unit_registry.load_definitions(path)



"""The code in this submodule is taken directly from the MIT-licensed
'spherical' package. Below is the original license text:

Copyright (c) 2021 Mike Boyle

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import math

import numpy as np


def A(j, j2, j3, m1):
    return math.sqrt(
        (j**2 - (j2 - j3) ** 2) * ((j2 + j3 + 1) ** 2 - j**2) * (j**2 - m1**2)
    )


def B(j, j2, j3, m2, m3):
    return (2 * j + 1) * (
        (m2 + m3) * (j2 * (j2 + 1) - j3 * (j3 + 1)) - (m2 - m3) * j * (j + 1)
    )


def Xf(j, j2, j3, m1):
    return j * A(j + 1, j2, j3, m1)


Yf = B


def Zf(j, j2, j3, m1):
    return (j + 1) * A(j, j2, j3, m1)


def normalize(f, j_min, j_max):
    norm = 0.0
    for j in range(j_min, j_max + 1):
        norm += (2 * j + 1) * f[j] ** 2
    f[j_min : j_max + 1] /= math.sqrt(norm)


def determine_signs(f, j_min, j_max, j2, j3, m2, m3):
    if (f[j_max] < 0.0 and (-1) ** (j2 - j3 + m2 + m3) > 0) or (
        f[j_max] > 0.0 and (-1) ** (j2 - j3 + m2 + m3) < 0
    ):
        f[j_min : j_max + 1] *= -1.0


class Wigner3jCalculator:
    def __init__(self, j2_max, j3_max):
        self._size = j2_max + j3_max + 1
        self.workspace = np.zeros(4 * self._size, dtype=np.float64)

    @property
    def size(self):
        return self._size

    def calculate(self, j2, j3, m2, m3):
        # Compute Wigner 3-j symbols
        #
        # For given values of j₂, j₃, m₂, m₃, this computes all valid values of
        #
        #     ⎛j₁  j₂  j₃⎞   _   ⎛j₂  j₃  j₁⎞   _   ⎛j₃  j₁  j₂⎞
        #     ⎝m₁  m₂  m₃⎠   -   ⎝m₂  m₃  m₁⎠   -   ⎝m₃  m₁  m₂⎠
        #
        # The valid values have m₁=-m₂-m₃ and j₁ ranging from max(|j₂-j₃|, |m₁|) to
        # j₂+j₃.
        # The calculation uses the approach described by Luscombe and Luban (1998)
        # <https://doi.org/10.1103/PhysRevE.57.7274>, which is a recurrence method,
        # leading to significant gains in speed and accuracy.
        #
        # The returned array is a slice of this object's `workspace` array, and so will
        # not
        # remain the same between calls to this function.  If you want to keep a copy of
        # the results, explicitly call the `numpy.copy` method.
        #
        # The returned array is indexed by j₁.  In particular, note that some invalid
        # j₁ indices are accessible, but have value 0.0.
        #
        # This implementation uses several tricks gleaned from the Fortran code in
        # <https://github.com/SHTOOLS/SHTOOLS>, which also implements the Luscombe-Luban
        # algorithm.  In particular, that code (and now this code) treats several special
        # cases that were not clearly specified by Luscombe-Luban.
        #
        # To use this object, do something like this:
        #
        # ..code-block:: python
        #
        #     # Try to do this just once because it allocates memory, which is slow
        #     calculator = Wigner3jCalculator(j2_max, j3_max)
        #
        #     # Presumably, the following is inside some loop over j2, j3, m2, m3
        #     w3j = calculator.calculate(j2, j3, m2, m3)
        #     m1 = - m2 - m3
        #     for j1 in range(max(abs(j2-j3), abs(m1)), j2+j3+1):
        #         w3j[j1]  # This is the value of the 3-j symbol written above
        #
        # Again, the array w3j contains accessible memory outside of the bounds of j1
        # given in the above loop, but those values will all be 0.0.

        m1 = -(m2 + m3)

        undefined_min = False
        undefined_max = False
        scale_factor = 1_000.0

        # Set up the workspace
        self.workspace[:] = 0.0
        f = self.workspace[: self.size]
        sf = self.workspace[self.size : 2 * self.size]
        rf = sf  # An alias for the same memory as `sf`
        F_minus = self.workspace[2 * self.size : 3 * self.size]
        F_plus = self.workspace[3 * self.size : 4 * self.size]

        # Calculate some useful bounds
        j_min = max(abs(j2 - j3), abs(m2 + m3))
        j_max = j2 + j3

        # Skip all calculations if there are no nonzero elements
        if abs(m2) > j2 or abs(m3) > j3:
            return f
        if j_max < j_min:
            return f

        # When only one term is present, we have a simple formula
        if j_max == j_min:
            f[j_min] = 1.0 / math.sqrt(2.0 * j_min + 1.0)
            if (f[j_min] < 0.0 and (-1) ** (j2 - j3 + m2 + m3) > 0) or (
                f[j_min] > 0.0 and (-1) ** (j2 - j3 + m2 + m3) < 0
            ):
                f[j_min] *= -1.0
            return f

        ##########################################################################
        # Forward iteration over first "nonclassical" region j_min ≤ j ≤ j_minus #
        ##########################################################################
        Xf_j_min = Xf(j_min, j2, j3, m1)
        Yf_j_min = Yf(j_min, j2, j3, m2, m3)

        if m1 == 0 and m2 == 0 and m3 == 0:
            # Recurrence is undefined, but it's okay because all odd terms must be zero
            F_minus[j_min] = 1.0
            F_minus[j_min + 1] = 0.0
            j_minus = j_min + 1

        elif Yf_j_min == 0.0:
            # The second term is either undefined or zero
            if Xf_j_min == 0.0:
                undefined_min = True
                j_minus = j_min
            else:
                F_minus[j_min] = 1.0
                F_minus[j_min + 1] = 0.0
                j_minus = j_min + 1

        elif Xf_j_min * Yf_j_min >= 0.0:
            # The second term is already in the classical region
            F_minus[j_min] = 1.0
            F_minus[j_min + 1] = -Yf_j_min / Xf_j_min
            j_minus = j_min + 1

        else:
            # Use recurrence relation, Eq. (3) from Luscombe and Luban (1998)
            sf[j_min] = -Xf_j_min / Yf_j_min
            j_minus = j_max
            for j in range(j_min + 1, j_max):
                denominator = Yf(j, j2, j3, m2, m3) + Zf(j, j2, j3, m1) * sf[j - 1]
                Xf_j = Xf(j, j2, j3, m1)
                if (
                    abs(Xf_j) > abs(denominator)
                    or Xf_j * denominator >= 0.0
                    or denominator == 0.0
                ):
                    j_minus = j - 1
                    break
                else:
                    sf[j] = -Xf_j / denominator
            F_minus[j_minus] = 1.0
            for k in range(1, j_minus - j_min + 1):
                F_minus[j_minus - k] = F_minus[j_minus - k + 1] * sf[j_minus - k]
            if j_minus == j_min:
                # Calculate at least two terms so that these can be used in three-term
                # recursion
                F_minus[j_min + 1] = -Yf_j_min / Xf_j_min
                j_minus = j_min + 1

        if j_minus == j_max:
            # We're finished!
            f[j_min : j_max + 1] = F_minus[j_min : j_max + 1]
            normalize(f, j_min, j_max)
            determine_signs(f, j_min, j_max, j2, j3, m2, m3)
            return f

        ##########################################################################
        # Reverse iteration over second "nonclassical" region j_plus ≤ j ≤ j_max #
        ##########################################################################
        Yf_j_max = Yf(j_max, j2, j3, m2, m3)
        Zf_j_max = Zf(j_max, j2, j3, m1)

        if m1 == 0 and m2 == 0 and m3 == 0:
            # Recurrence is undefined, but it's okay because all odd terms must be zero
            F_plus[j_max] = 1.0
            F_plus[j_max - 1] = 0.0
            j_plus = j_max - 1

        elif Yf_j_max == 0.0:
            # The second term is either undefined or zero
            if Zf_j_max == 0.0:
                undefined_max = True
                j_plus = j_max
            else:
                F_plus[j_max] = 1.0
                F_plus[j_max - 1] = -Yf_j_max / Zf_j_max
                j_plus = j_max - 1

        elif Yf_j_max * Zf_j_max >= 0.0:
            # The second term is already in the classical region
            F_plus[j_max] = 1.0
            F_plus[j_max - 1] = -Yf_j_max / Zf_j_max
            j_plus = j_max - 1

        else:
            # Use recurrence relation, Eq. (2) from Luscombe and Luban (1998)
            rf[j_max] = -Zf_j_max / Yf_j_max
            j_plus = j_min
            for j in range(j_max - 1, j_minus - 1, -1):
                denominator = Yf(j, j2, j3, m2, m3) + Xf(j, j2, j3, m1) * rf[j + 1]
                Zf_j = Zf(j, j2, j3, m1)
                if (
                    denominator == 0.0
                    or abs(Zf_j) > abs(denominator)
                    or Zf_j * denominator >= 0.0
                ):
                    j_plus = j + 1
                    break
                else:
                    rf[j] = -Zf_j / denominator
            F_plus[j_plus] = 1.0
            for k in range(1, j_max - j_plus + 1):
                F_plus[j_plus + k] = F_plus[j_plus + k - 1] * rf[j_plus + k]
            if j_plus == j_max:
                F_plus[j_max - 1] = -Yf_j_max / Zf_j_max
                j_plus = j_max - 1

        ######################################################################
        # Three-term recurrence over "classical" region j_minus ≤ j ≤ j_plus #
        ######################################################################
        if undefined_min and undefined_max:
            raise ValueError(
                "Cannot initialize recurrence in Wigner3jCalculator.calculate"
            )

        if not undefined_min and not undefined_max:
            # Iterate upwards and downwards, meeting in the middle
            j_mid = (j_minus + j_plus) // 2
            for j in range(j_minus, j_mid):
                F_minus[j + 1] = -(
                    Yf(j, j2, j3, m2, m3) * F_minus[j]
                    + Zf(j, j2, j3, m1) * F_minus[j - 1]
                ) / Xf(j, j2, j3, m1)
                if abs(F_minus[j + 1]) > 1.0:
                    F_minus[j_min : j + 1 + 1] /= scale_factor
                if abs(F_minus[j + 1] / F_minus[j - 1]) < 1.0 and F_minus[j + 1] != 0.0:
                    j_mid = j + 1
                    break
            F_minus_j_mid = F_minus[j_mid]
            if (
                F_minus[j_mid - 1] != 0.0
                and abs(F_minus_j_mid / F_minus[j_mid - 1]) < 1.0e-6
            ):
                j_mid -= 1
                F_minus_j_mid = F_minus[j_mid]
            for j in range(j_plus, j_mid, -1):
                F_plus[j - 1] = -(
                    Xf(j, j2, j3, m1) * F_plus[j + 1]
                    + Yf(j, j2, j3, m2, m3) * F_plus[j]
                ) / Zf(j, j2, j3, m1)
                if abs(F_plus[j - 1]) > 1.0:
                    F_plus[j - 1 : j_max + 1] /= scale_factor
            F_plus_j_mid = F_plus[j_mid]
            if j_mid == j_max:
                f[j_min : j_max + 1] = F_minus[j_min : j_max + 1]
            elif j_mid == j_min:
                f[j_min : j_max + 1] = F_plus[j_min : j_max + 1]
            else:
                f[j_min : j_mid + 1] = (
                    F_minus[j_min : j_mid + 1] * F_plus_j_mid / F_minus_j_mid
                )
                f[j_mid + 1 : j_max + 1] = F_plus[j_mid + 1 : j_max + 1]

        elif not undefined_min and undefined_max:
            # Iterate upwards only
            for j in range(j_minus, j_plus):
                F_minus[j + 1] = -(
                    Zf(j, j2, j3, m1) * F_minus[j - 1]
                    + Yf(j, j2, j3, m2, m3) * F_minus[j]
                ) / Xf(j, j2, j3, m1)
                if abs(F_minus[j + 1]) > 1:
                    F_minus[j_min : j + 1 + 1] /= scale_factor
            f[j_min : j_max + 1] = F_minus[j_min : j_max + 1]

        elif undefined_min and not undefined_max:
            # Iterate downwards only
            for j in range(j_plus, j_min, -1):
                F_plus[j - 1] = -(
                    Xf(j, j2, j3, m1) * F_plus[j + 1]
                    + Yf(j, j2, j3, m2, m3) * F_plus[j]
                ) / Zf(j, j2, j3, m1)
                if abs(F_plus[j - 1]) > 1:
                    F_plus[j - 1 : j_max + 1] /= scale_factor
            f[j_min : j_max + 1] = F_plus[j_min : j_max + 1]

        #############
        # Finish up #
        #############
        normalize(f, j_min, j_max)
        determine_signs(f, j_min, j_max, j2, j3, m2, m3)
        return f


def Wigner3j(j_1, j_2, j_3, m_1, m_2, m_3):
    if m_1 + m_2 + m_3 != 0:
        return 0.0
    if abs(m_1) > j_1 or abs(m_2) > j_2 or abs(m_3) > j_3:
        return 0.0
    # Permute cyclically to ensure that j_1 is the largest
    if j_1 == max(j_1, j_2, j_3):
        pass
    elif j_2 == max(j_1, j_2, j_3):
        j_1, j_2, j_3 = j_2, j_3, j_1
        m_1, m_2, m_3 = m_2, m_3, m_1
    else:  # j_3 == max(j_1, j_2, j_3)
        j_1, j_2, j_3 = j_3, j_1, j_2
        m_1, m_2, m_3 = m_3, m_1, m_2
    if j_1 > j_2 + j_3:
        return 0.0
    calculator = Wigner3jCalculator(j_2, j_3)
    w3j = calculator.calculate(j_2, j_3, m_2, m_3)
    return w3j[j_1]


def clebsch_gordan(j_1, m_1, j_2, m_2, j_3, m_3):
    return (
        (-1.0) ** (j_1 - j_2 + m_3)
        * math.sqrt(2 * j_3 + 1)
        * Wigner3j(j_1, j_2, j_3, m_1, m_2, -m_3)
    )



__all__ = ["core", "light_curves", "orbits", "units"]

from jaxoplanet import (
    core as core,
    light_curves as light_curves,
    orbits as orbits,
    units as units,
)
from jaxoplanet.jaxoplanet_version import __version__ as __version__



import pytest

from jaxoplanet.starry.ylm import Ylm
from jaxoplanet.test_utils import assert_allclose


@pytest.mark.parametrize("data", [{(5, 0): 5.0, (3, -1): 1.5}])
def test_starry_ylm_compare(data):
    starry = pytest.importorskip("starry")
    theano = pytest.importorskip("theano")
    theano.config.gcc__cxxflags += " -fexceptions"
    starry.config.lazy = False

    data = dict(data)
    data.update({(0, 0): 1.0})
    ylm = Ylm(data)
    starry_map = starry.Map(ylm.deg)
    for (ell, m), c in data.items():
        if (ell, m) == (0, 0):
            continue
        starry_map[ell, m] = c
    expected = starry_map._y
    calc = ylm.todense()
    assert_allclose(expected, calc)
